"""Plugin lint command."""

from __future__ import annotations

import io
import logging
import zipfile
from pathlib import Path

import requests
import rich_click as click
from pydantic import ValidationError

from hcli.lib.console import console
from hcli.lib.ida.plugin import (
    IDAMetadataDescriptor,
    parse_plugin_version,
    validate_metadata_in_plugin_archive,
)
from hcli.lib.ida.plugin.install import validate_metadata_in_plugin_directory
from hcli.lib.ida.plugin.repo import fetch_plugin_archive
from hcli.lib.util.logging import m

logger = logging.getLogger(__name__)


def _lint_readme_in_directory(plugin_path: Path, source_name: str) -> int:
    """Check for README.md file in plugin directory.

    returns: number of recommendations made
    """
    found_files = []
    has_exact_match = False
    recommendation_count = 0

    for item in plugin_path.iterdir():
        if item.is_file() and item.name == "README.md":
            has_exact_match = True
            break
        if item.is_file() and item.name.lower().startswith("readme"):
            found_files.append(item.name)

    if has_exact_match:
        return 0

    if found_files:
        console.print(
            f"[yellow]Recommendation[/yellow] ({source_name}): rename {found_files[0]} to README.md (exact casing)"
        )
        console.print("  Use 'README.md' with exact casing for consistency and discoverability")
        recommendation_count += 1
    else:
        console.print(f"[yellow]Recommendation[/yellow] ({source_name}): add a README.md file")
        console.print("  A README helps users understand what your plugin does and how to use it")
        recommendation_count += 1

    return recommendation_count


def _lint_readme_in_archive(zip_data: bytes, metadata_path: Path, source_name: str) -> int:
    """Check for README.md file in plugin archive.

    returns: number of recommendations made
    """
    plugin_dir = metadata_path.parent
    recommendation_count = 0

    with zipfile.ZipFile(io.BytesIO(zip_data), "r") as zip_file:
        namelist = zip_file.namelist()

        found_files = []
        has_exact_match = False

        for file_path in namelist:
            path_obj = Path(file_path)
            if path_obj.parent == plugin_dir:
                if path_obj.name == "README.md":
                    has_exact_match = True
                    break
                if path_obj.name.lower().startswith("readme"):
                    found_files.append(path_obj.name)

        if has_exact_match:
            return 0

        if found_files:
            console.print(
                f"[yellow]Recommendation[/yellow] ({source_name}): rename {found_files[0]} to README.md (exact casing)"
            )
            console.print("  Use 'README.md' with exact casing for consistency and discoverability")
            recommendation_count += 1
        else:
            console.print(f"[yellow]Recommendation[/yellow] ({source_name}): add a README.md file")
            console.print("  A README helps users understand what your plugin does and how to use it")
            recommendation_count += 1

        return recommendation_count


def _check_unexpected_keys(metadata: IDAMetadataDescriptor, source_name: str) -> int:
    """Check for unexpected keys in the plugin metadata.

    returns: number of recommendations made
    """
    recommendation_count = 0

    if hasattr(metadata.plugin, "__pydantic_extra__") and metadata.plugin.__pydantic_extra__:
        extra_keys = sorted(metadata.plugin.__pydantic_extra__.keys())
        for key in extra_keys:
            console.print(f"[yellow]Warning[/yellow] ({source_name}): unexpected key in plugin metadata: '{key}'")
            console.print("  This key is not part of the ida-plugin.json schema and will be ignored")
            recommendation_count += 1

    return recommendation_count


def _lint_metadata(metadata: IDAMetadataDescriptor, source_name: str) -> int:
    """Validate a single plugin metadata and show lint recommendations.

    returns: number of recommendations made
    """
    recommendation_count = 0

    recommendation_count += _check_unexpected_keys(metadata, source_name)

    if not parse_plugin_version(metadata.plugin.version):
        console.print(f"[red]Error[/red] ({source_name}): plugin version should look like 'X.Y.Z'")
        recommendation_count += 1

    if not metadata.plugin.ida_versions:
        console.print(f"[yellow]Recommendation[/yellow] ({source_name}): ida-plugin.json: provide plugin.idaVersions")
        console.print("  Specify which IDA versions your plugin supports (e.g., ['9.0', '9.1'])")
        recommendation_count += 1

    if not metadata.plugin.description:
        console.print(f"[yellow]Recommendation[/yellow] ({source_name}): ida-plugin.json: provide plugin.description")
        console.print("  A one-line description improves discoverability in the plugin repository")
        recommendation_count += 1

    if not metadata.plugin.categories:
        console.print(f"[yellow]Recommendation[/yellow] ({source_name}): ida-plugin.json: provide plugin.categories")
        console.print("  Categories help users find your plugin (e.g., 'malware-analysis', 'decompilation')")
        recommendation_count += 1

    if not metadata.plugin.logo_path:
        console.print(f"[yellow]Recommendation[/yellow] ({source_name}): ida-plugin.json: provide plugin.logoPath")
        console.print("  A logo image (16:9 aspect ratio) makes your plugin more visually appealing")
        recommendation_count += 1

    if not metadata.plugin.keywords:
        console.print(f"[yellow]Recommendation[/yellow] ({source_name}): ida-plugin.json: provide plugin.keywords")
        console.print("  Keywords improve search discoverability in the plugin repository")
        recommendation_count += 1

    if not metadata.plugin.license:
        console.print(f"[yellow]Recommendation[/yellow] ({source_name}): ida-plugin.json: provide plugin.license")
        console.print("  Specify the license (e.g., 'MIT', 'Apache 2.0') to clarify usage rights")
        recommendation_count += 1

    if not metadata.plugin.authors and not metadata.plugin.maintainers:
        console.print(
            f"[red]Error[/red] ({source_name}): ida-plugin.json: provide plugin.authors or plugin.maintainers"
        )
        console.print("  Contact information is required for authors or maintainers")
        recommendation_count += 1
    else:
        # Check if contacts have both name and email for better completeness
        for i, author in enumerate(metadata.plugin.authors):
            if not author.name:
                console.print(
                    f"[yellow]Recommendation[/yellow] ({source_name}): plugin.authors[{i}]: provide an author name"
                )
                recommendation_count += 1
        for i, maintainer in enumerate(metadata.plugin.maintainers):
            if not maintainer.name:
                console.print(
                    f"[yellow]Recommendation[/yellow] ({source_name}): plugin.maintainers[{i}]: provide a maintainer name"
                )
                recommendation_count += 1

    return recommendation_count


def _lint_plugin_directory(plugin_path: Path) -> int:
    """Lint a plugin in a directory.

    returns: number of recommendations made
    """
    recommendation_count = 0

    metadata_file = plugin_path / "ida-plugin.json"
    if not metadata_file.exists():
        console.print(f"[red]Error[/red]: ida-plugin.json not found in {plugin_path}")
        return 1

    content = metadata_file.read_text(encoding="utf-8")
    try:
        metadata = IDAMetadataDescriptor.model_validate_json(content)
    except ValidationError as e:
        console.print("[red]Error[/red]: ida-plugin.json validation failed")
        for error in e.errors():
            field_path = ".".join(str(loc) for loc in error["loc"])
            error_msg = error["msg"]
            error_type = error["type"]

            if error_type == "missing":
                console.print(f"  [red]Missing required field[/red]: {field_path}")
            else:
                console.print(f"  [red]Invalid value[/red] for {field_path}: {error_msg}")

            recommendation_count += 1

        return recommendation_count

    try:
        validate_metadata_in_plugin_directory(plugin_path)
    except Exception as e:
        console.print(f"[red]Error[/red]: ida-plugin.json validation failed: {e}")
        recommendation_count += 1
        return recommendation_count

    recommendation_count += _lint_metadata(metadata, str(plugin_path))
    recommendation_count += _lint_readme_in_directory(plugin_path, str(plugin_path))

    return recommendation_count


def _lint_plugin_archive(zip_data: bytes, source_name: str) -> int:
    """Lint plugins in a .zip archive from bytes.

    returns: number of recommendations made
    """
    recommendation_count = 0

    plugins_found = []
    with zipfile.ZipFile(io.BytesIO(zip_data), "r") as zip_file:
        for file_path in zip_file.namelist():
            if not file_path.endswith("ida-plugin.json"):
                continue

            logger.debug(m("found metadata path: %s", file_path))
            with zip_file.open(file_path) as f:
                try:
                    metadata = IDAMetadataDescriptor.model_validate_json(f.read().decode("utf-8"))
                except ValidationError as e:
                    logger.debug(m("failed to validate metadata: %s", file_path, path=file_path, error=str(e)))
                    console.print(f"[red]Error[/red] ({source_name}): {file_path}: ida-plugin.json validation failed")
                    for error in e.errors():
                        field_path = ".".join(str(loc) for loc in error["loc"])
                        error_msg = error["msg"]
                        error_type = error["type"]

                        if error_type == "missing":
                            console.print(f"  [red]Missing required field[/red]: {field_path}")
                        else:
                            console.print(f"  [red]Invalid value[/red] for {field_path}: {error_msg}")
                        recommendation_count += 1
                    continue
                else:
                    logger.debug(m("found valid metadata: %s", file_path))
                    plugins_found.append((Path(file_path), metadata))

    for path, meta in plugins_found:
        logger.debug("found plugin %s at %s", meta.plugin.name, path)

    if not plugins_found:
        console.print(f"[red]Error[/red]: No valid plugins found in archive {source_name}")
        recommendation_count += 1
        return recommendation_count

    for metadata_path, metadata in plugins_found:
        plugin_source_name = f"{source_name}:{metadata_path}"

        try:
            validate_metadata_in_plugin_archive(zip_data, metadata_path, metadata)
        except ValidationError as e:
            console.print(
                f"[red]Error[/red] ({plugin_source_name}): {metadata_path}: ida-plugin.json validation failed"
            )
            for error in e.errors():
                field_path = ".".join(str(loc) for loc in error["loc"])
                error_msg = error["msg"]
                error_type = error["type"]

                if error_type == "missing":
                    console.print(f"  [red]Missing required field[/red]: {field_path}")
                else:
                    console.print(f"  [red]Invalid value[/red] for {field_path}: {error_msg}")
                recommendation_count += 1

            continue

        except Exception as e:
            console.print(f"[red]Error[/red]: {metadata_path}: ida-plugin.json validation failed: {e}")
            recommendation_count += 1
            continue

        recommendation_count += _lint_metadata(metadata, plugin_source_name)
        recommendation_count += _lint_readme_in_archive(zip_data, metadata_path, plugin_source_name)

    return recommendation_count


@click.command()
@click.argument(
    "path",
    metavar="PATH|URL",
)
def lint_plugin_directory(path: str) -> None:
    """Lint an IDA plugin directory, archive (.zip file), or HTTPS URL."""
    recommendation_count = 0
    if path.startswith("https://"):
        logger.info("linting from HTTP URL")
        try:
            buf = fetch_plugin_archive(path)
        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
            console.print(f"[red]Cannot connect to {path} - network unavailable.[/red]")
            console.print("Please check your internet connection.")
            raise click.Abort()
        except Exception as e:
            console.print(f"[red]Error[/red]: Failed to fetch archive from {path}: {e}")
            raise click.Abort()

        recommendation_count = _lint_plugin_archive(buf, path)

    else:
        plugin_path = Path(path).expanduser().resolve()
        if not plugin_path.exists():
            console.print(f"[red]Error[/red]: Path does not exist: {plugin_path}")
            raise click.Abort()

        if plugin_path.is_file():
            if plugin_path.suffix.lower() != ".zip":
                console.print(f"[red]Error[/red]: File must be a .zip archive: {plugin_path}")
                raise click.Abort()

            zip_data = plugin_path.read_bytes()
            recommendation_count = _lint_plugin_archive(zip_data, str(plugin_path))
        elif plugin_path.is_dir():
            recommendation_count = _lint_plugin_directory(plugin_path)
        else:
            console.print(f"[red]Error[/red]: Path must be a directory or .zip file: {plugin_path}")
            raise click.Abort()

    if not recommendation_count:
        console.print("[green]no recommendations[/green]")



"""IDA Pro utilities for installation and path management."""

import errno
import json
import logging
import os
import re
import shutil
import stat
import subprocess
import tempfile
from dataclasses import dataclass
from functools import total_ordering
from pathlib import Path
from typing import Literal, NamedTuple

import rich.console
from pydantic import BaseModel, ConfigDict, Field

from hcli.env import ENV
from hcli.lib.util.cache import get_cache_directory
from hcli.lib.util.io import NoSpaceError, check_free_space, get_os

logger = logging.getLogger(__name__)


class DownloadResource(NamedTuple):
    """IDA download resource information."""

    id: str
    name: str
    description: str
    category: str
    version: str
    os: str
    arch: str


@dataclass
@total_ordering
class IdaProduct:
    product: str
    major: int
    minor: int
    suffix: str | None = None

    @classmethod
    def from_installer_filename(cls, filename: str):
        """Parse IDA installer filename to extract version information.

        Args:
            filename: IDA installer filename (e.g., 'ida-pro_92_x64linux.run')

        Raises:
            ValueError: If filename format is not recognized
        """
        basename = filename
        for ext in [".app.zip", ".run", ".exe"]:
            if basename.endswith(ext):
                basename = basename[: -len(ext)]
                break

        # filename pattern: ida-{product}_{version}_{platform}
        match = re.match(r"^ida-([^_]+)_(\d{2})(sp\d+)?_", basename)
        if not match:
            raise ValueError(f"Unrecognized installer filename format: {filename}")

        product_part = match.group(1)  # like: pro, home-pc, essential
        version_major = int(match.group(2)[0])  # like: 9
        version_minor = int(match.group(2)[1])  # like: 1
        suffix = match.group(3) if match.group(3) else None  # like: sp1

        product_mapping = {
            "pro": "IDA Professional",
            "home-pc": "IDA Home",
            "home-arm": "IDA Home",
            "home-mips": "IDA Home",
            "home-ppc": "IDA Home",
            "home-riscv": "IDA Home",
            "free-pc": "IDA Free",
            "essential": "IDA Essential",
            "classroom-free": "IDA Classroom",
        }

        product = product_mapping.get(product_part, f"IDA {product_part.title()}")
        return cls(product, version_major, version_minor, suffix)

    def __str__(self):
        base = f"{self.product} {self.major}.{self.minor}"
        return f"{base}{self.suffix}" if self.suffix else base

    def __lt__(self, other):
        if not isinstance(other, IdaProduct):
            return NotImplemented
        return (self.product, self.major, self.minor, self.suffix or "") < (
            other.product,
            other.major,
            other.minor,
            other.suffix or "",
        )


def is_installable(download: DownloadResource) -> bool:
    """Check if a download resource is installable on the current platform."""
    current_os = get_os()
    src = download.id

    return (
        (src.endswith(".app.zip") and current_os == "mac")
        or (src.endswith(".run") and current_os == "linux")
        or (src.endswith(".exe") and current_os == "windows")
    )


def get_ida_user_dir() -> Path:
    """Get the IDA Pro user directory."""
    # duplicate here, because we prefer access through ENV
    # but tests might update env vars for the current process.
    idausr = os.environ.get("HCLI_IDAUSR")
    if idausr:
        return Path(idausr)
    if ENV.HCLI_IDAUSR is not None:
        return Path(ENV.HCLI_IDAUSR)

    if ENV.IDAUSR is not None:
        return Path(ENV.IDAUSR)

    os_ = get_os()
    if os_ == "windows":
        appdata = os.environ.get("APPDATA")
        if not appdata:
            raise ValueError("Failed to determine %APPDATA% location: environment variable not set")

        return Path(appdata) / "Hex-Rays" / "IDA Pro"
    elif os_ in ("linux", "mac"):
        home = os.environ.get("HOME")
        if not home:
            raise ValueError("Failed to determine home directory: environment variable not set")
        return Path(home) / ".idapro"
    else:
        raise ValueError(f"Unsupported operating system: {os_}")


def get_user_home_dir() -> Path:
    """Get the user home directory."""
    os_ = get_os()
    if os_ == "windows":
        appdata = os.environ.get("APPDATA")
        if not appdata:
            raise ValueError("Failed to determine %APPDATA% location: environment variable not set")

        return Path(appdata)
    elif os_ in ("linux", "mac"):
        home = os.environ.get("HOME")
        if not home:
            raise ValueError("Failed to determine home directory: environment variable not set")
        return Path(home)
    else:
        raise ValueError(f"Unsupported operating system: {os_}")


def get_default_ida_install_directory(ver: IdaProduct) -> Path:
    """Get the default installation directory for IDA Pro."""

    # like "IDA Professional 9.1sp1"
    app_directory_name = str(ver)

    os_ = get_os()
    if os_ == "windows":
        return Path(os.environ.get("ProgramFiles", r"C:\Program Files")) / app_directory_name
    elif os_ == "linux":
        # workaround for #99: idat from IDA 9.2 on Linux fails to start if the path contains a space.
        # so we avoid using the path component "IDA Professional 9.2" and instead use "IDA-Professional-9.2"
        # which is ugly but works.
        #
        # typically idat isn't widely used; however, HCLI does use it to discover the path to IDA's Python interpreter,
        # as well as the installed arch (ARM or Intel on macOS). The latter could probably be discovered by inspecting
        # the installed files; however, figuring out the Python configuration is messy, and much easier to leave to idat.
        #
        # see also the warnings in commands/ida/install.py.
        if ver.major == 9 and ver.minor == 2 and " " in app_directory_name:
            # "IDA Professional 9.2" -> "IDA-Professional-9.2"
            sanitized_name = app_directory_name.replace(" ", "-")
            logger.info(
                f"Sanitized installation directory name for IDA 9.2 on Linux: '{app_directory_name}' -> '{sanitized_name}'"
            )
            app_directory_name = sanitized_name

        return get_user_home_dir() / ".local" / "share" / "applications" / app_directory_name
    elif os_ == "mac":
        return Path("/Applications/") / f"{app_directory_name}.app"
    else:
        raise ValueError(f"Unsupported operating system: {os_}")


def get_ida_path(ida_dir: Path) -> Path:
    """Get the IDA application path from the installation directory."""
    if get_os() == "mac":
        return Path(ida_dir) / "Contents" / "MacOS"
    else:
        return Path(ida_dir)


def get_ida_binary_path(ida_dir: Path, suffix: str = "") -> Path:
    """Get the IDA binary path."""
    if get_os() == "windows":
        return Path(get_ida_path(ida_dir)) / f"ida{suffix}.exe"
    else:
        return Path(get_ida_path(ida_dir)) / f"ida{suffix}"


def get_idat_path(ida_dir: Path) -> Path:
    """Get the IDA text-mode (idat) executable path."""
    return get_ida_binary_path(ida_dir, "t")


def find_standard_windows_installations() -> list[Path]:
    """Find standard IDA Pro installations on Windows."""
    ret = []

    base_directory = Path(os.environ.get("ProgramFiles", r"C:\Program Files"))

    # Check the base directory for IDA installations
    if base_directory.exists():
        for entry in base_directory.iterdir():
            if not entry.is_dir():
                continue

            if not entry.name.startswith("IDA Pro"):
                continue

            ret.append(entry)

    return ret


def find_standard_linux_installations() -> list[Path]:
    """Find standard IDA Pro installations on Linux."""
    # TODO: can also look in registered XDG applications, or maybe in /opt
    ret = []
    base_directory = get_user_home_dir() / ".local" / "share" / "applications"

    if base_directory.exists():
        for entry in base_directory.iterdir():
            if not entry.is_dir():
                continue

            if not entry.name.startswith("IDA Pro"):
                continue

            ret.append(entry)

    return ret


def find_standard_mac_installations() -> list[Path]:
    """Find standard IDA Pro installations on macOS."""
    ret = []

    base_directory = Path("/Applications")

    # Check the base directory for IDA installations
    if base_directory.exists():
        for entry in base_directory.iterdir():
            if not entry.is_dir():
                continue

            if not entry.name.startswith("IDA Pro"):
                continue

            ret.append(entry)

    return ret


def find_standard_installations() -> list[Path]:
    """Find standard IDA Pro installations."""
    ret = set()

    try:
        ret.add(find_current_ida_install_directory())
    except ValueError:
        pass

    os_ = get_os()
    if os_ == "windows":
        ret.update(find_standard_windows_installations())
    elif os_ == "linux":
        ret.update(find_standard_linux_installations())
    elif os_ == "mac":
        ret.update(find_standard_mac_installations())
    else:
        raise ValueError(f"Unsupported operating system: {os_}")

    return list(ret)


def is_ida_dir(ida_dir: Path) -> bool:
    """Check if a directory contains a valid IDA installation."""
    binary_path = Path(get_ida_binary_path(ida_dir))
    return binary_path.exists()


def install_license(license_path: Path, target_path: Path) -> None:
    """Install a license file to an IDA directory."""
    target_file = target_path / license_path.name
    shutil.copy2(license_path, target_file)


def get_license_dir(ida_dir: Path) -> Path:
    """Get the license directory for an IDA installation."""
    return get_ida_path(ida_dir)


def accept_eula(install_dir: Path) -> None:
    # Accept the EULA (to be persistent across runs - you need to mount $HOME/.idapro as a volume)
    os.environ["IDADIR"] = str(install_dir)
    try:
        # force this to be imported first and not reordered by ruff
        import idapro  # noqa: F401
        import ida_registry  # isort: skip
    except Exception:
        raise RuntimeError("idalib not available")

    ida_registry.reg_write_int("EULA 90", 1)
    ida_registry.reg_write_int("EULA 91", 1)
    ida_registry.reg_write_int("EULA 92", 1)
    logger.info("EULA accepted")


def install_ida(installer: Path, install_dir: Path):
    """
    Install IDA Pro from an installer.

    Args:
      installer: path to the installer downloaded from the Hex-Rays portal.
      install_dir: path to the installation directory, which should not already exist.

    Installation directory should look like:
      - %Program Files%\\IDA Professional 9.1\\
      - /Applications/IDA Professional 9.1.app/
      - /opt/ida-9.1/
      - /tmp/ida-9.1/
      - ...
    """
    if install_dir.exists():
        raise FileExistsError(
            f"Installation directory already exists: {install_dir}\n"
            f"Please remove the existing directory first or choose a different location."
        )

    logger.info(f"Installing IDA in {install_dir}")

    installer_size = installer.stat().st_size
    check_free_space(install_dir.parent, installer_size * 3)

    install_dir.mkdir(parents=True, exist_ok=False)

    try:
        current_os = get_os()
        if current_os == "mac":
            _install_ida_mac(installer, install_dir)
        elif current_os == "linux":
            _install_ida_unix(installer, install_dir)
        elif current_os == "windows":
            _install_ida_windows(installer, install_dir)
        else:
            raise ValueError(f"unsupported OS: {current_os}")
    except Exception as e:
        logger.error(f"Installation failed: {e}")
        raise

    contents = list(install_dir.iterdir())
    logger.debug("installed contents: %s", contents)
    if not len(contents):
        raise RuntimeError("installation failed: installation directory contents not created")

    has_ida_hlp = False
    for _, _, files in os.walk(install_dir):
        if "ida.hlp" in files:
            has_ida_hlp = True

    if not has_ida_hlp:
        raise RuntimeError("installation failed: ida.hlp not created")


def _install_ida_mac(installer: Path, prefix: Path) -> None:
    """Install IDA on macOS."""
    if not shutil.which("unzip"):
        raise RuntimeError("unzip is required to install IDA on macOS")

    with tempfile.TemporaryDirectory(prefix="hcli_") as temp_unpack_dir:
        with tempfile.TemporaryDirectory(prefix="hcli_") as temp_install_dir:
            logger.info(f"Unpacking installer to {temp_unpack_dir}...")

            # Unpack the installer
            process = subprocess.run(["unzip", "-qq", str(installer), "-d", temp_unpack_dir], capture_output=True)

            if process.returncode != 0:
                raise RuntimeError("Failed to unpack installer")

            entries = list(Path(temp_unpack_dir).iterdir())
            if len(entries) != 1:
                raise ValueError(f"unexpected contents of zip archive: {len(entries)} root directories")

            # typically this is the app name, like `ida-pro_90_armmac.app`
            # however the directory name might not be precisely the same as the zip archive filename
            # such as in SP releases.
            app_name = entries[0]

            installer_path = None
            for platform in ("osx-arm64", "osx-x86_64"):
                candidate_path = Path(temp_unpack_dir) / app_name / "Contents" / "MacOS" / platform
                if candidate_path.exists():
                    installer_path = candidate_path
                    break

            if not installer_path:
                raise RuntimeError("Installer executable not found")

            logger.info(f"Running installer {app_name}...")
            temp_install_path = Path(temp_install_dir)
            args = _get_installer_args(temp_install_path)

            process = subprocess.run([str(installer_path)] + args, capture_output=True)

            if process.returncode != 0:
                raise RuntimeError("Installer execution failed")

            # Find installed folder and copy to prefix
            installed_folders = list(temp_install_path.iterdir())

            if not installed_folders:
                raise RuntimeError("No installation found after running installer")

            install_folder = installed_folders[0]
            _copy_dir(install_folder, prefix)


def _install_ida_unix(installer: Path, prefix: Path) -> None:
    """Install IDA on Unix/Linux."""
    args = _get_installer_args(prefix)

    installer_path = Path(installer)

    # If installer is not absolute and has no directory component, prefix with './'
    if not installer_path.is_absolute() and installer_path.parent == Path("."):
        installer_path = Path(f"./{installer_path}")

    if not os.access(installer_path, os.X_OK):
        logger.info(f"Setting executable permission on {installer_path}")
        current_mode = os.stat(installer_path).st_mode
        os.chmod(installer_path, current_mode | stat.S_IXUSR)

    home_dir = get_user_home_dir()
    share_dir = Path(home_dir) / ".local" / "share" / "applications"
    share_dir.mkdir(parents=True, exist_ok=True)

    process = subprocess.run([str(installer_path)] + args, capture_output=True)

    if process.returncode != 0:
        raise RuntimeError("Installer execution failed")


def _install_ida_windows(installer: Path, prefix: Path) -> None:
    """Install IDA on Windows."""
    args = _get_installer_args(prefix)

    process = subprocess.run(["cmd", "/c", str(installer)] + args, capture_output=True)

    if process.returncode != 0:
        raise RuntimeError("Installer execution failed")


def _get_installer_args(prefix: Path) -> list[str]:
    """Get installer arguments."""
    args = ["--mode", "unattended", "--debugtrace", "debug.log"]

    if get_os() == "windows":
        args.extend(["--install_python", "0"])

    if prefix:
        args.extend(["--prefix", str(prefix)])

    return args


def _copy_dir(src_path: Path, dest_path: Path) -> None:
    """Copy directory recursively."""
    if not src_path.exists():
        return

    dest_path.mkdir(parents=True, exist_ok=True)

    for item in src_path.rglob("*"):
        relative_path = item.relative_to(src_path)
        dest_item = dest_path / relative_path

        if item.is_dir():
            dest_item.mkdir(parents=True, exist_ok=True)
        elif item.is_file():
            dest_item.parent.mkdir(parents=True, exist_ok=True)
            try:
                shutil.copy2(item, dest_item)
            except OSError as e:
                if e.errno == errno.ENOSPC:
                    if dest_item.exists():
                        dest_item.unlink()
                    raise NoSpaceError(dest_path) from e
                raise


class PathsConfig(BaseModel):
    model_config = ConfigDict(serialize_by_alias=True)  # type: ignore

    # like: "/Applications/IDA Professional 9.1.app/Contents/MacOS"
    installation_directory: Path | None = Field(alias="ida-install-dir", default=None)


class PluginRepositoryConfig(BaseModel):
    model_config = ConfigDict(serialize_by_alias=True)  # type: ignore

    url: str = Field(
        default="https://raw.githubusercontent.com/HexRaysSA/plugin-repository/refs/heads/v1/plugin-repository.json",
    )


class SettingsConfig(BaseModel):
    model_config = ConfigDict(serialize_by_alias=True)  # type: ignore

    plugin_repository: PluginRepositoryConfig = Field(
        alias="plugin-repository", default_factory=lambda: PluginRepositoryConfig()
    )


class PluginConfig(BaseModel):
    # `ida-plugin.json` `.plugin.settings` describes the schema for these settings.
    settings: dict[str, str | bool] = Field(default_factory=dict)


# describes contents of IDAUSR/ida-config.json
class IDAConfigJson(BaseModel):
    """IDA configuration $IDAUSR/ida-config.json"""

    model_config = ConfigDict(serialize_by_alias=True)  # type: ignore

    version: Literal[1] | None = Field(alias="Version", default=1)
    paths: PathsConfig = Field(alias="Paths", default_factory=lambda: PathsConfig())
    settings: SettingsConfig = Field(alias="Settings", default_factory=lambda: SettingsConfig())
    # from plugin name to config.
    plugins: dict[str, PluginConfig] = Field(alias="Plugins", default_factory=dict)


def get_ida_config_path() -> Path:
    idausr = get_ida_user_dir()

    return Path(idausr) / "ida-config.json"


def get_ida_config() -> IDAConfigJson:
    ida_config_path = get_ida_config_path()
    if not ida_config_path.exists():
        logger.debug("using default ida-config.json contents")
        return IDAConfigJson()

    return IDAConfigJson.model_validate_json(ida_config_path.read_text(encoding="utf-8"))


def set_ida_config(config: IDAConfigJson):
    ida_config_path = get_ida_config_path()
    if not ida_config_path.exists():
        logger.debug("creating $IDAUSR directory")
        ida_config_path.parent.mkdir(parents=True, exist_ok=True)

    _ = ida_config_path.write_text(config.model_dump_json())


class MissingCurrentInstallationDirectory(ValueError):
    def __init__(self, msg):
        super().__init__(f"failed to determine current IDA installation directory: {msg}")


class FailedToDetectIDAVersion(RuntimeError):
    def __init__(self, msg: str | None = None):
        if msg:
            super().__init__(f"failed to determine current IDA version: {msg}")
        else:
            super().__init__("failed to determine current IDA version")


def find_current_ida_install_directory() -> Path:
    # duplicate here, because we prefer access through ENV
    # but tests might update env vars for the current process.
    env = os.environ.get("HCLI_CURRENT_IDA_INSTALL_DIR")
    if env:
        return Path(env)
    if ENV.HCLI_CURRENT_IDA_INSTALL_DIR is not None:
        return Path(ENV.HCLI_CURRENT_IDA_INSTALL_DIR)

    if ENV.IDADIR is not None:
        return Path(ENV.IDADIR)

    config = get_ida_config()
    if not config.paths.installation_directory:
        raise MissingCurrentInstallationDirectory("directory doesn't exist")

    if not config.paths.installation_directory.exists():
        raise MissingCurrentInstallationDirectory("ida-config.json invalid: ida-install-dir doesn't exist")

    logger.debug("current IDA installation: %s", config.paths.installation_directory)
    return config.paths.installation_directory


def explain_missing_current_installation_directory(console: rich.console.Console):
    console.print("[red]Error[/red]: failed to find the current IDA Pro installation directory.")
    console.print("")
    console.print("You can configure this in two ways:")
    console.print("")
    console.print("1. set the default value in $IDAUSR/ida-config.json, which you can do via:")
    console.print("")
    console.print("     [grey69]hcli ida set-default /path/to/IDA/installation/[/grey69]")
    console.print("")
    console.print("2. provide the HCLI_CURRENT_IDA_INSTALL_DIR environment variable, like:")
    console.print("")
    console.print("     [grey69]export HCLI_CURRENT_IDA_INSTALL_DIR=/path/to/IDA/installation/[/grey69] # Linux, or")
    console.print(
        '     [grey69]export HCLI_CURRENT_IDA_INSTALL_DIR="/Applications/IDA Professional 9.2.app/Contents/MacOS/"[/grey69] # macOS, or'
    )
    console.print(
        '     [grey69]set HCLI_CURRENT_IDA_INSTALL_DIR="C:\\Program Files\\IDA Professional 9.2"[/grey69]  # Windows'
    )
    console.print("")


def explain_failed_to_detect_ida_version(console: rich.console.Console):
    console.print("[red]Error[/red]: failed to determine current IDA version.")
    console.print("")
    console.print("hcli needs to run IDA to detect its version, but this failed.")
    console.print("")
    console.print("You can work around this by explicitly setting the version:")
    console.print("")
    console.print("1. set the HCLI_CURRENT_IDA_VERSION environment variable:")
    console.print("")
    console.print("     [grey69]export HCLI_CURRENT_IDA_VERSION=9.2[/grey69]")
    console.print("")
    console.print("2. also ensure IDA installation directory is configured:")
    console.print("")
    console.print("     [grey69]hcli ida set-default /path/to/IDA/installation/[/grey69]")
    console.print("")
    console.print("   or via environment variable:")
    console.print("")
    console.print("     [grey69]export HCLI_CURRENT_IDA_INSTALL_DIR=/path/to/IDA/installation/[/grey69]")
    console.print("")


def find_current_idat_executable() -> Path:
    install_directory = find_current_ida_install_directory()
    current_os = get_os()
    if current_os == "mac" and install_directory.name == "MacOS":
        # strip off trailing Contents/MacOS
        install_directory = install_directory.parent.parent

    return get_idat_path(install_directory)


def run_py_in_current_idapython(src: str) -> str:
    idat_path = find_current_idat_executable()
    if not idat_path.exists():
        raise ValueError(f"can't find idat: {idat_path}")

    if get_os() == "linux" and "9.2" in str(idat_path.absolute()) and " " in str(idat_path.absolute()):
        logger.warning(
            "invoking idat on IDA 9.2/Linux with a space in the full path, you might encounter HCLI GitHub issue #99"
        )

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)

        script_path = temp_path / "idat-script.py"
        log_path = temp_path / "ida.log"

        script_path.write_text(src)

        # invoke like:
        #
        #     idat -a -A -c -t -L"/absolute/path/to/ida.log" -S"/absolute/path/to/idat-script.py"
        #
        # -a disable auto analysis
        # -A autuonomous, no dialogs
        # -c delete old database
        # -t create an empty database
        # -L"/absolute/path/to/ida.log"
        # -S"/absolute/path/to/script.py"
        cmd = [
            str(idat_path),
            "-a",  # disable auto analysis
            "-A",  # autonomous, no dialogs
            "-c",  # delete old database
            "-t",  # create an empty database
            f"-L{str(log_path.absolute())}",
            f"-S{str(script_path.absolute())}",
        ]

        result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        logger.debug(f"idat command: {' '.join(cmd)}")
        logger.debug(f"idat exit code: {result.returncode}")
        if result.stdout:
            logger.debug(f"idat stdout: {result.stdout}")
        if result.stderr:
            logger.debug(f"idat stderr: {result.stderr}")

        if not log_path.exists():
            raise RuntimeError(f"failed to invoke idat: log file was not created: {log_path}")

        for line in log_path.read_text().splitlines():
            if not line.startswith("__hcli__:"):
                continue

            return json.loads(line[len("__hcli__:") :])

        raise RuntimeError("failed to invoke idat: could not find expected lines in log output")


def get_current_ida_platform_cache_path() -> Path:
    return get_cache_directory("current-ida") / "platform.json"


def set_current_ida_platform_cache(ida_path: Path, platform: str) -> None:
    cache_path = get_current_ida_platform_cache_path()
    if cache_path.exists():
        doc = json.loads(cache_path.read_text(encoding="utf-8"))
    else:
        doc = {}
    doc[str(ida_path.absolute())] = platform
    cache_path.write_text(json.dumps(doc), encoding="utf-8")


def get_current_ida_platform_cache(ida_path: Path) -> str:
    cache_path = get_current_ida_platform_cache_path()
    if not cache_path.exists():
        raise KeyError(f"No platform cache found for {ida_path}")

    doc = json.loads(cache_path.read_text(encoding="utf-8"))
    return doc[str(ida_path.absolute())]


FIND_PLATFORM_PY = """
# output like:
#
#     __hcli__:"windows-x86_64"
import sys
import json
import platform

system = platform.system()
if system == "Windows":
    plat = "windows-x86_64"
elif system == "Linux":
    plat = "linux-x86_64"
elif system == "Darwin":
    # via: https://stackoverflow.com/questions/7491391/
    version = platform.uname().version
    if "RELEASE_ARM64" in version:
        plat = "macos-aarch64"
    elif "RELEASE_X86_64" in version:
        plat = "macos-x86_64"
    else:
        raise ValueError(f"Unsupported macOS version: {version}")
else:
    raise ValueError(f"Unsupported OS: {system}")
print("__hcli__:" + json.dumps(plat))
sys.exit()
"""


def find_current_ida_platform() -> str:
    """find the platform associated with the current IDA installation"""
    # duplicate here, because we prefer access through ENV
    # but tests might update env vars for the current process.
    env = os.environ.get("HCLI_CURRENT_IDA_PLATFORM")
    if env:
        return env
    if ENV.HCLI_CURRENT_IDA_PLATFORM is not None:
        return ENV.HCLI_CURRENT_IDA_PLATFORM

    os_ = get_os()
    if os_ == "windows":
        return "windows-x86_64"
    elif os_ == "linux":
        return "linux-x86_64"
    elif os_ == "mac":
        ida_dir = find_current_ida_install_directory()
        try:
            return get_current_ida_platform_cache(ida_dir)
        except KeyError:
            pass
        try:
            platform = run_py_in_current_idapython(FIND_PLATFORM_PY)
        except RuntimeError as e:
            raise RuntimeError("failed to determine current IDA platform") from e
        set_current_ida_platform_cache(ida_dir, platform)
        return platform
    else:
        raise ValueError(f"Unsupported OS: {os_}")


def get_current_ida_version_cache_path() -> Path:
    return get_cache_directory("current-ida") / "version.json"


def set_current_ida_version_cache(ida_path: Path, version: str) -> None:
    cache_path = get_current_ida_version_cache_path()
    if cache_path.exists():
        doc = json.loads(cache_path.read_text(encoding="utf-8"))
    else:
        doc = {}
    doc[str(ida_path.absolute())] = version
    cache_path.write_text(json.dumps(doc), encoding="utf-8")


def get_current_ida_version_cache(ida_path: Path) -> str:
    cache_path = get_current_ida_version_cache_path()
    if not cache_path.exists():
        raise KeyError(f"No version cache found for {ida_path}")

    doc = json.loads(cache_path.read_text(encoding="utf-8"))
    return doc[str(ida_path.absolute())]


FIND_VERSION_PY = """
# output like:
#
#     __hcli__:"9.1"
import sys
import json
import ida_kernwin
print("__hcli__:" + json.dumps(ida_kernwin.get_kernel_version()))
sys.exit()
"""


def find_current_ida_version() -> str:
    """find the version of the current IDA installation, like '9.1'"""
    # duplicate here, because we prefer access through ENV
    # but tests might update env vars for the current process.
    env = os.environ.get("HCLI_CURRENT_IDA_VERSION")
    if env:
        return env
    if ENV.HCLI_CURRENT_IDA_VERSION is not None:
        return ENV.HCLI_CURRENT_IDA_VERSION

    ida_dir = find_current_ida_install_directory()
    try:
        return get_current_ida_version_cache(ida_dir)
    except KeyError:
        pass
    try:
        version = run_py_in_current_idapython(FIND_VERSION_PY)
    except RuntimeError as e:
        raise FailedToDetectIDAVersion() from e
    set_current_ida_version_cache(ida_dir, version)
    return version


def generate_instance_name(path: Path) -> str:
    """Generate a reasonable instance name from installation path."""
    # For macOS: "IDA Professional 9.2.app" -> "ida-pro-9.2"
    # For others: "IDA Professional 9.2" -> "ida-pro-9.2"
    name = path.name

    # Remove .app extension for macOS
    if name.endswith(".app"):
        name = name[:-4]

    # Convert to lowercase and replace spaces with dashes
    name = name.lower().replace(" ", "-")

    # Shorten common patterns
    name = name.replace("ida-professional", "ida-pro")
    name = name.replace("ida-home", "ida-home")
    name = name.replace("ida-free", "ida-free")

    return name


def add_instance_to_config(name: str, path: Path) -> bool:
    """Add an IDA instance to the configuration.

    Returns:
        True if the instance was added, False if it already exists.
    """
    from hcli.lib.config import config_store

    # Get existing instances
    instances: dict[str, str] = config_store.get_object("ke.ida.instances", {}) or {}

    if name in instances:
        return False  # Already exists

    # Store the absolute path as string
    instances[name] = str(path.absolute())

    # Save back to config
    config_store.set_object("ke.ida.instances", instances)

    return True



from __future__ import annotations

import rich_click as click
from rich.table import Table

from hcli.lib.console import console


def collect_all_commands(group: click.Group, parent_path: str = "") -> list[str]:
    """Recursively collect all command paths from a Click group."""
    commands = []

    for name, command in group.commands.items():
        current_path = f"{parent_path} {name}".strip()

        if isinstance(command, click.Group):
            # It's a group, recurse into it
            commands.extend(collect_all_commands(command, current_path))
        else:
            # It's a command, add it to the list
            commands.append(current_path)

    return commands


@click.command(name="commands")
@click.pass_context
def commands(ctx: click.Context):
    """List all available command combinations."""
    root_group = ctx.find_root().command
    if not isinstance(root_group, click.Group):
        console.print("[red]Error: Root command is not a group[/red]")
        return
    all_commands = collect_all_commands(root_group)

    table = Table(title="All Available Commands", show_header=True, header_style="bold blue")
    table.add_column("Command", style="green")
    table.add_column("Description", style="dim")

    for cmd_path in sorted(all_commands):
        # Try to get the command's help text
        try:
            parts = cmd_path.split()
            current_group: click.Command = root_group

            # Navigate to the command
            for part in parts[:-1]:
                if isinstance(current_group, click.Group):
                    current_group = current_group.commands[part]
                else:
                    raise AttributeError("Not a group")

            if isinstance(current_group, click.Group):
                command = current_group.commands[parts[-1]]
            else:
                raise AttributeError("Not a group")
            help_text = command.help or "No description available"
            # Get only the first line of the help text
            help_text = help_text.split("\n")[0].strip()

        except (KeyError, AttributeError):
            help_text = "No description available"

        table.add_row(f"hcli {cmd_path}", help_text)

    console.print(table)
    console.print(f"\n[dim]Total commands: {len(all_commands)}[/dim]")



import inspect
import logging
from pathlib import Path

from hcli.lib.ida import PluginConfig, get_ida_config, set_ida_config
from hcli.lib.ida.plugin import ChoiceValueError, PluginSettingDescriptor
from hcli.lib.ida.plugin.install import get_metadata_from_plugin_directory, get_plugin_directory, get_plugins_directory

logger = logging.getLogger(__name__)


def parse_setting_value(descriptor: PluginSettingDescriptor, string_value: str) -> str | bool:
    """Parse a string value to the appropriate type based on the setting descriptor.

    Args:
        descriptor: the setting descriptor
        string_value: the string value to parse

    Returns:
        The parsed value (str or bool)
    """
    if descriptor.type == "boolean":
        if string_value.lower() == "true":
            return True
        elif string_value.lower() == "false":
            return False
        else:
            raise ValueError(f"mismatching settings types: expected boolean ('true' or 'false'), got '{string_value}'")
    elif descriptor.type == "string":
        return string_value
    else:
        raise ValueError(f"unsupported setting type: {descriptor.type}")


def set_plugin_setting(plugin_name: str, key: str, value: str | bool):
    plugin_path = get_plugin_directory(plugin_name)
    metadata = get_metadata_from_plugin_directory(plugin_path)
    descr = metadata.plugin.get_setting(key)

    if descr.type == "string" and not isinstance(value, str):
        raise ValueError(f"mismatching settings types: {plugin_name}: {key}: {descr.type} vs {type(value).__name__}")
    elif descr.type == "boolean" and not isinstance(value, bool):
        raise ValueError(f"mismatching settings types: {plugin_name}: {key}: {descr.type} vs {type(value).__name__}")

    try:
        descr.validate_value(value)
    except ChoiceValueError as e:
        choices_str = ", ".join(e.choices)
        raise ValueError(
            f"failed to validate setting value: {plugin_name}: {key}: '{value}' (must be one of: {choices_str})"
        ) from e
    except ValueError as e:
        raise ValueError(f"failed to validate setting value: {plugin_name}: {key}: '{value}'") from e

    config = get_ida_config()
    if plugin_name not in config.plugins:
        plugin_config = PluginConfig()
    else:
        plugin_config = config.plugins[plugin_name]

    if plugin_config.settings.get(key) == value:
        return

    plugin_config.settings[key] = value
    config.plugins[plugin_name] = plugin_config

    set_ida_config(config)


def get_plugin_setting(plugin_name: str, key: str) -> str | bool:
    plugin_path = get_plugin_directory(plugin_name)
    metadata = get_metadata_from_plugin_directory(plugin_path)
    descr = metadata.plugin.get_setting(key)

    config = get_ida_config()
    if plugin_name not in config.plugins:
        if descr.default is not None:
            return descr.default
        else:
            raise KeyError(f"plugin setting not found: {plugin_name}: {key}")

    plugin_config = config.plugins[plugin_name]
    if key not in plugin_config.settings:
        if descr.default is not None:
            return descr.default
        else:
            raise KeyError(f"plugin setting not found: {plugin_name}: {key}")

    value = plugin_config.settings[key]
    try:
        descr.validate_value(value)
    except ChoiceValueError as e:
        choices_str = ", ".join(e.choices)
        raise ValueError(
            f"failed to validate existing setting value: {plugin_name}: {key}: '{value}' (must be one of: {choices_str})"
        ) from e
    except ValueError as e:
        raise ValueError(f"failed to validate existing setting value: {plugin_name}: {key}: '{value}'") from e

    return value


def del_plugin_setting(plugin_name: str, key: str):
    plugin_path = get_plugin_directory(plugin_name)
    metadata = get_metadata_from_plugin_directory(plugin_path)
    descr = metadata.plugin.get_setting(key)

    if descr.required and not descr.default:
        raise ValueError(f"cannot delete required setting without default: {plugin_name}: {key}")

    config = get_ida_config()
    if plugin_name not in config.plugins:
        raise KeyError(f"plugin setting not found: {plugin_name}: {key}")

    plugin_config = config.plugins[plugin_name]
    if key not in plugin_config.settings:
        raise KeyError(f"plugin setting not found: {plugin_name}: {key}")

    del plugin_config.settings[key]
    config.plugins[plugin_name] = plugin_config

    set_ida_config(config)


def has_plugin_setting(plugin_name: str, key: str) -> bool:
    """Check if a plugin setting is explicitly set.

    Args:
        plugin_name: the plugin name
        key: the setting key

    Returns: True if the setting is explicitly set, False otherwise
    """
    plugin_path = get_plugin_directory(plugin_name)
    metadata = get_metadata_from_plugin_directory(plugin_path)
    metadata.plugin.get_setting(key)

    config = get_ida_config()
    if plugin_name not in config.plugins:
        return False

    plugin_config = config.plugins[plugin_name]
    return key in plugin_config.settings


def get_current_plugin() -> str:
    """Get the plugin name by walking the call stack.

    This must only be called from IDA Pro plugins, or it will raise RuntimeError.

    Returns:
        The plugin name extracted from the first plugin module found in the call stack.
    """

    # to determine the current plugin, we'll walk the call stack
    # searching for modules that seem to be an IDA plugin.
    #
    # while we'd prefer to use the module name, which is set to `__plugins__<name>`,
    # this doesn't work when we're not in the entrypoint of a module
    # (such as a registered plugmod_t thats not found in the entrypoint python file).
    #
    # so, we fall back to also checking the file system path to see if it
    # falls within the plugins directory ($IDAUSR/plugins/).
    # this also helps when the plugin name has been normalized from "my-foo" to "__plugins__my_foo".

    frame = inspect.currentframe()
    if frame is None:
        raise RuntimeError("failed to get current frame")

    plugins_path = str(get_plugins_directory().absolute())

    current_frame = frame.f_back
    while current_frame is not None:
        logger.debug("inspecting frame: %s", current_frame)
        module_name = current_frame.f_globals.get("__name__")
        module_filename = current_frame.f_code.co_filename
        if module_filename.startswith(plugins_path):
            # check file path first, because it handles normalization better

            module_relative_path = Path(module_filename).relative_to(plugins_path)
            plugin_directory_name = module_relative_path.parts[0]

            plugin_directory = Path(plugins_path) / plugin_directory_name
            metadata = get_metadata_from_plugin_directory(plugin_directory)
            plugin_name = metadata.plugin.name
            logger.debug("found plugin by path: %s %s", module_filename, plugin_name)
            return plugin_name
        elif module_name and module_name.startswith("__plugins__"):
            plugin_name = module_name[len("__plugins__") :]
            logger.debug("found plugin by module name: %s %s", module_name, plugin_name)
            return plugin_name

        current_frame = current_frame.f_back

    raise RuntimeError("get_current_plugin() must be called from within a plugin module")


def get_current_plugin_setting(key: str) -> str | bool:
    plugin = get_current_plugin()
    return get_plugin_setting(plugin, key)


def set_current_plugin_setting(key: str, value: str | bool):
    plugin = get_current_plugin()
    return set_plugin_setting(plugin, key, value)


def del_current_plugin_setting(key: str):
    plugin = get_current_plugin()
    return del_plugin_setting(plugin, key)


def has_current_plugin_setting(key: str) -> bool:
    plugin = get_current_plugin()
    return has_plugin_setting(plugin, key)


def list_current_plugin_settings() -> list[PluginSettingDescriptor]:
    """List all setting descriptors for the current plugin.

    This must only be called from IDA Pro plugins, or it will raise RuntimeError.

    Returns:
        List of PluginSettingDescriptor instances defined for the current plugin.
    """
    plugin = get_current_plugin()
    plugin_path = get_plugin_directory(plugin)
    metadata = get_metadata_from_plugin_directory(plugin_path)
    return metadata.plugin.settings



"""Common license utilities and functions."""

from __future__ import annotations

from datetime import datetime, timezone
from pathlib import Path
from typing import Callable

import questionary

from hcli.commands.common import safe_ask_async
from hcli.lib.api.license import License, license
from hcli.lib.console import console
from hcli.lib.constants import cli


async def select_licenses(customer_id: str, predicate: Callable[[License], bool] | None = None) -> list[License]:
    """
    Select licenses interactively or return all matching licenses.

    Args:
        customer_id: Customer ID to get licenses for
        predicate: Optional filter function for licenses

    Returns:
        List of selected licenses
    """
    if predicate is None:

        def predicate(lic):
            return True

    licenses = await license.get_licenses(customer_id)
    filtered = [lic for lic in licenses if predicate(lic)]

    if len(filtered) == 1:
        return filtered
    elif len(filtered) == 0:
        console.print("[yellow]No licenses found matching criteria[/yellow]")
        return []
    else:
        # Group licenses by product catalog
        legacy = [lic for lic in filtered if lic.product_catalog == "legacy"]
        subscription = [lic for lic in filtered if lic.product_catalog == "subscription"]
        all_licenses = legacy + subscription

        # Create choices for questionary
        choices: list[questionary.Choice] = []

        # Add legacy licenses
        if legacy:
            choices.append(questionary.Separator("Perpetual licenses:"))
            for lic in legacy:
                choices.append(questionary.Choice(license_to_string(lic), lic))

        # Add subscription licenses
        if subscription:
            choices.append(questionary.Separator("Subscription licenses:"))
            for lic in subscription:
                choices.append(questionary.Choice(license_to_string(lic), lic))

        # Use questionary for selection
        selected = await safe_ask_async(
            questionary.checkbox(
                "Select licenses:",
                choices=choices,
                style=cli.CHECKBOX_STYLE,
            )
        )

        if not selected:
            return []

        if "all" in selected:
            return all_licenses

        return selected


async def download_license(
    customer_id: str,
    lic: License,
    target_dir: str,
    ask_assets: bool = True,
) -> list[str]:
    """
    Download a single license.

    Args:
        customer_id: Customer ID
        lic: License to download
        target_dir: Target directory for downloads
        ask_assets: Whether to ask which asset types to download

    Returns:
        List of downloaded file paths
    """
    results: list[str] = []
    asset_types = lic.asset_types

    if not asset_types:
        console.print("[yellow]This license has no assets to download.[/yellow]")
        return results

    if ask_assets and len(asset_types) > 0:
        asset_type = asset_types[0]
        if len(asset_types) > 1:
            console.print(f"\n[bold]Available asset types for license {lic.pubhash}:[/bold]")
            for i, asset in enumerate(asset_types, 1):
                console.print(f"  {i}. {asset}")

            selection = await safe_ask_async(
                questionary.select(
                    "Select asset type:",
                    choices=[questionary.Choice(f"{i}. {asset}", str(i)) for i, asset in enumerate(asset_types, 1)],
                    default="1",
                    style=cli.SELECT_STYLE,
                )
            )
            asset_type = asset_types[int(selection) - 1]

        result = await download_license_asset(customer_id, lic, asset_type, target_dir)
        if result:
            results.append(result)
    else:
        # Download all asset types
        for asset_type in asset_types:
            result = await download_license_asset(customer_id, lic, asset_type, target_dir)
            if result:
                results.append(result)

    return results


async def download_license_asset(
    customer_id: str,
    lic: License,
    asset_type: str,
    target_dir: str,
) -> str | None:
    """
    Download a specific license asset.

    Args:
        customer_id: Customer ID
        lic: License object
        asset_type: Type of asset to download
        target_dir: Target directory for download

    Returns:
        Downloaded file path or None if failed
    """
    try:
        if not lic.pubhash:
            console.print(f"[red]License has no ID for asset {asset_type}[/red]")
            return None
        filename = await license.download_license(customer_id, lic.pubhash, asset_type, target_dir)
        if filename:
            console.print(f"[green]License {asset_type} for {lic.pubhash} downloaded as: {filename}[/green]")
            return filename
        else:
            console.print("[red]Failed to download license[/red]")
            return None
    except Exception as e:
        console.print(f"[red]Error downloading license: {e}[/red]")
        return None


def license_to_string(lic: License) -> str:
    """
    Convert license object to human-readable string.

    Args:
        lic: License object

    Returns:
        Formatted license string
    """
    text = "does not expire"

    if lic.end_date:
        try:
            end_date = datetime.fromisoformat(lic.end_date.replace("Z", "+00:00"))
            now = datetime.now(timezone.utc)

            if end_date < now:
                # Calculate time since expiration
                delta = now - end_date
                if delta.days > 365:
                    years = delta.days // 365
                    text = f"expired {years} year{'s' if years != 1 else ''} ago"
                elif delta.days > 30:
                    months = delta.days // 30
                    text = f"expired {months} month{'s' if months != 1 else ''} ago"
                elif delta.days > 0:
                    text = f"expired {delta.days} day{'s' if delta.days != 1 else ''} ago"
                else:
                    text = "expired today"
            else:
                # Calculate time until expiration
                delta = end_date - now
                if delta.days > 365:
                    years = delta.days // 365
                    text = f"expires in {years} year{'s' if years != 1 else ''}"
                elif delta.days > 30:
                    months = delta.days // 30
                    text = f"expires in {months} month{'s' if months != 1 else ''}"
                elif delta.days > 0:
                    text = f"expires in {delta.days} day{'s' if delta.days != 1 else ''}"
                else:
                    text = "expires today"
        except (ValueError, TypeError):
            # Fallback if date parsing fails
            text = f"expires {lic.end_date}"

    # Get decompilers and other addons
    decompilers = [
        addon.product.code
        for addon in (lic.addons or [])
        if addon.product and addon.product.product_subtype == "DECOMPILER"
    ]
    other = [
        addon.product.code
        for addon in (lic.addons or [])
        if addon.product and addon.product.product_subtype != "DECOMPILER"
    ]

    suffix = ""
    if decompilers:
        suffix += f"{len(decompilers)} decompiler{'s' if len(decompilers) != 1 else ''} "
    if other:
        suffix += f"[{', '.join(other)}]"

    edition_name = lic.edition.edition_name if lic.edition else "Unknown"
    return f"{lic.pubhash} {edition_name} [{lic.license_type}] {text} {suffix}".strip()


def ensure_target_directory(target_dir: str) -> str:
    """
    Ensure target directory exists and return the absolute path.

    Args:
        target_dir: Target directory path

    Returns:
        Absolute path to the target directory
    """
    path = Path(target_dir).expanduser().resolve()
    path.mkdir(parents=True, exist_ok=True)
    return str(path)



"""License API client."""

from pydantic import BaseModel

from .common import get_api_client


class Product(BaseModel):
    """Product information."""

    id: int
    code: str
    name: str
    family: str | None = None
    catalog: str
    edition: str | None = None
    platform: str | None = None
    ui_label: str | None = None
    base_code: str | None = None
    update_code: str | None = None
    product_type: str
    product_subtype: str | None = None


class Addon(BaseModel):
    """License addon information."""

    id: int | None = None
    pubhash: str | None = None
    license_key: str | None = None
    seats: int | None = None
    password: str | None = None
    start_date: str | None = None
    end_date: str | None = None
    product_code: str | None = None
    created_at: str | None = None
    updated_at: str | None = None
    activation_owner: str | None = None
    activation_owner_type: str | None = None
    product: "Product | None" = None


class Edition(BaseModel):
    """Edition information."""

    id: int | None = None
    tags: list[str] | None = None
    plan_id: str | None = None
    max_items: int | None = None
    plan_name: str | None = None
    edition_id: str | None = None
    edition_name: str | None = None
    build_edition_id: str | None = None
    build_product_id: str | None = None


class License(BaseModel):
    """License information."""

    id: int | None = None
    pubhash: str | None = None
    plan_id: str | None = None
    license_key: str | None = None
    start_date: str | None = None
    end_date: str | None = None
    license_type: str | None = None
    seats: int | None = None
    password: str | None = None
    product_code: str | None = None
    customer_id: int | None = None
    end_customer_id: int | None = None
    activation_owner: str | None = None
    activation_owner_type: str | None = None
    status: str | None = None
    comment: str | None = None
    created_at: str | None = None
    updated_at: str | None = None
    generation_date: str | None = None
    download_date: str | None = None
    addons: "list[Addon] | None" = None
    edition: "Edition | None" = None
    asset_types: list[str] | None = None
    format_version: str | None = None
    product_catalog: str | None = None
    end_customer_visible: bool | None = None


class PagedResponse(BaseModel):
    """Paged response wrapper."""

    items: list[License]
    total: int


class LicenseAPI:
    """License API client."""

    async def get_licenses(self, customer_id: str) -> list[License]:
        """Get licenses for a customer."""
        client = await get_api_client()
        data = await client.get_json(f"/api/licenses/{customer_id}?page=1&limit=100")
        paged_response = PagedResponse(**data)

        # Sort licenses by end_date (descending, with null dates last)
        licenses = paged_response.items
        licenses.sort(key=lambda x: (x.end_date is None, x.end_date), reverse=True)

        return licenses

    async def download_license(
        self, customer_id: str, license_id: str, asset_type: str, target_dir: str = "./"
    ) -> str | None:
        """Download a license file."""
        client = await get_api_client()
        download_url = await client.get_json(f"/api/licenses/{customer_id}/download/{asset_type}/{license_id}")

        if download_url:
            return await client.download_file(download_url, target_dir=target_dir, auth=False)

        return None


# Global instance
license = LicenseAPI()



from __future__ import annotations

import rich_click as click
from rich.table import Table

from hcli.commands.common import select_customer
from hcli.lib.api.license import license
from hcli.lib.commands import async_command, auth_command
from hcli.lib.console import console


@click.option(
    "-p",
    "--plan",
    type=click.Choice(["subscription", "legacy"]),
    help="Filter by plan type: subscription or legacy",
)
@auth_command()
@async_command
async def list_licenses(plan: str | None) -> None:
    """List available licenses with rich formatting."""
    # Select customer
    customer_obj = await select_customer()
    if not customer_obj:
        return

    # Get licenses
    if not customer_obj.id:
        console.print("[red]Customer ID not available[/red]")
        return
    licenses = await license.get_licenses(str(customer_obj.id))

    if not licenses:
        console.print("[yellow]No licenses found[/yellow]")
        return

    # Filter by plan if specified
    if plan:
        licenses = [lic for lic in licenses if lic.product_catalog == plan]
        if not licenses:
            console.print(f"[yellow]No {plan} licenses found[/yellow]")
            return

    # Group licenses by product catalog
    legacy_licenses = [lic for lic in licenses if lic.product_catalog == "legacy"]
    subscription_licenses = [lic for lic in licenses if lic.product_catalog == "subscription"]

    console.print(f"\n[bold]Licenses for customer [{customer_obj.id}]:[/bold]")

    # Display legacy licenses
    if legacy_licenses:
        console.print(f"\n[cyan bold]Perpetual Licenses ({len(legacy_licenses)}):[/cyan bold]")
        _display_licenses_table(legacy_licenses)

    # Display subscription licenses
    if subscription_licenses:
        console.print(f"\n[cyan bold]Subscription Licenses ({len(subscription_licenses)}):[/cyan bold]")
        _display_licenses_table(subscription_licenses)

    console.print(f"\n[dim]Total: {len(licenses)} license(s)[/dim]")


def _display_licenses_table(licenses) -> None:
    """Display licenses in a formatted table."""
    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("ID", style="cyan", no_wrap=True)
    table.add_column("Edition", style="green")
    table.add_column("Type", style="yellow")
    table.add_column("Status", style="blue")
    table.add_column("Expiration", style="red")
    table.add_column("Addons", style="magenta")

    for lic in licenses:
        # Format status with color
        if lic.status == "active":
            status = "[green]Active[/green]"
        elif lic.status == "expired":
            status = "[red]Expired[/red]"
        else:
            status = f"[yellow]{lic.status.title()}[/yellow]"

        # Format expiration
        expiration_text = "Never"
        if lic.end_date:
            try:
                from datetime import datetime, timezone

                end_date = datetime.fromisoformat(lic.end_date.replace("Z", "+00:00"))
                now = datetime.now(timezone.utc)

                if end_date < now:
                    delta = now - end_date
                    if delta.days > 365:
                        years = delta.days // 365
                        expiration_text = f"[red]{years}y ago[/red]"
                    elif delta.days > 30:
                        months = delta.days // 30
                        expiration_text = f"[red]{months}mo ago[/red]"
                    else:
                        expiration_text = f"[red]{delta.days}d ago[/red]"
                else:
                    delta = end_date - now
                    if delta.days > 365:
                        years = delta.days // 365
                        expiration_text = f"[yellow]{years}y[/yellow]"
                    elif delta.days > 30:
                        months = delta.days // 30
                        expiration_text = f"[yellow]{months}mo[/yellow]"
                    else:
                        expiration_text = f"[yellow]{delta.days}d[/yellow]"
            except (ValueError, TypeError):
                expiration_text = lic.end_date

        # Format addons
        decompilers = [addon.product.code for addon in lic.addons if addon.product.product_subtype == "DECOMPILER"]
        other = [addon.product.code for addon in lic.addons if addon.product.product_subtype != "DECOMPILER"]

        addon_parts = []
        if decompilers:
            addon_parts.append(f"{len(decompilers)} decompiler(s)")
        if other:
            addon_parts.append(", ".join(other))

        addons = " + ".join(addon_parts) if addon_parts else "None"

        table.add_row(
            lic.pubhash,
            lic.edition.edition_name if lic.edition else "Unknown",
            lic.license_type,
            status,
            expiration_text,
            addons,
        )

    console.print(table)



from __future__ import annotations

import rich_click as click


@click.group()
def ida() -> None:
    """Manage IDA installations."""
    pass


from .accept_eula import accept_eula_command  # noqa: E402
from .install import install  # noqa: E402
from .set_default import set_default_ida  # noqa: E402

ida.add_command(accept_eula_command)
ida.add_command(install)
ida.add_command(set_default_ida)



import json
import urllib.request
from pathlib import Path
from typing import Literal
from urllib.parse import urlparse

import requests
from pydantic import BaseModel

from hcli.lib.ida.plugin.repo import BasePluginRepo, Plugin


class StaticPluginRepo(BaseModel):
    version: Literal[1] = 1
    plugins: list[Plugin]


class JSONFilePluginRepo(BasePluginRepo):
    def __init__(self, plugins: list[Plugin]):
        super().__init__()
        self.plugins = plugins

    def get_plugins(self) -> list[Plugin]:
        return self.plugins

    def to_json(self):
        doc = StaticPluginRepo(plugins=self.get_plugins()).model_dump_json()
        # pydantic doesn't have a way to emit json with sorted keys
        # and we want a deterministic file,
        # so we re-encode here.
        return json.dumps(json.loads(doc), sort_keys=True, indent=4)

    def to_bytes(self) -> bytes:
        return self.to_json().encode("utf-8")

    @classmethod
    def from_json(cls, doc: str):
        return cls(StaticPluginRepo.model_validate_json(doc).plugins)

    @classmethod
    def from_bytes(cls, buf: bytes):
        return cls.from_json(buf.decode("utf-8"))

    @classmethod
    def from_file(cls, path: Path):
        return cls.from_bytes(path.read_bytes())

    @classmethod
    def from_url(cls, url: str):
        parsed_url = urlparse(url)

        if parsed_url.scheme == "file":
            file_path = Path(urllib.request.url2pathname(parsed_url.path))
            if not file_path.exists():
                raise FileNotFoundError(f"File not found: {file_path}")
            return cls.from_bytes(file_path.read_bytes())

        elif parsed_url.scheme == "https":
            response = requests.get(url, timeout=30.0)
            response.raise_for_status()
            return cls.from_bytes(response.content)

        else:
            raise ValueError(f"Unsupported URL scheme: {parsed_url.scheme}")

    @classmethod
    def from_repo(cls, other: BasePluginRepo):
        return cls(other.get_plugins())



"""Exceptions for IDA plugin installation and management."""

from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path


class PluginInstallationError(Exception):
    """Base exception for plugin installation failures."""

    pass


class PluginAlreadyInstalledError(PluginInstallationError):
    """Plugin is already installed."""

    def __init__(self, name: str, path: Path):
        self.name = name
        self.path = path
        super().__init__(
            f"Plugin '{name}' is already installed at {path}. "
            f"Use 'hcli plugin upgrade {name}' to update or 'hcli plugin uninstall {name}' first."
        )


class PlatformIncompatibleError(PluginInstallationError):
    """Current platform is not supported by the plugin."""

    def __init__(self, current: str, supported: Sequence[str]):
        self.current = current
        self.supported = supported
        platforms_str = ", ".join(sorted(supported))
        super().__init__(
            f"Plugin not compatible with current platform '{current}'. Supported platforms: {platforms_str}"
        )


class IDAVersionIncompatibleError(PluginInstallationError):
    """Current IDA version is not supported by the plugin."""

    def __init__(self, current: str, supported: Sequence[str]):
        self.current = current
        self.supported = supported
        # Show first 10 supported versions to avoid overwhelming output
        if len(supported) > 10:
            versions_str = ", ".join(supported[:10]) + f" (and {len(supported) - 10} more)"
        else:
            versions_str = ", ".join(supported)
        super().__init__(f"Plugin not compatible with IDA version '{current}'. Supported versions: {versions_str}")


class PipNotAvailableError(PluginInstallationError):
    """pip is not available in IDA's Python environment."""

    def __init__(self):
        super().__init__(
            "Cannot install plugin: pip is not available in IDA's Python environment. "
            "The plugin requires Python dependencies but pip cannot be found. "
            "Please ensure your IDA installation includes pip support."
        )


class DependencyInstallationError(PluginInstallationError):
    """Python dependencies cannot be installed."""

    def __init__(self, dependencies: Sequence[str], reason: str | None = None):
        self.dependencies = dependencies
        self.reason = reason
        deps_str = ", ".join(dependencies)
        msg = f"Cannot install required Python dependencies: {deps_str}"
        if reason:
            msg += f". Reason: {reason}"
        super().__init__(msg)


class InvalidPluginNameError(PluginInstallationError):
    """Plugin name is invalid."""

    def __init__(self, name: str, reason: str):
        self.name = name
        self.reason = reason
        super().__init__(f"Invalid plugin name '{name}': {reason}")


class PluginNotInstalledError(Exception):
    """Plugin is not installed."""

    def __init__(self, name: str):
        self.name = name
        super().__init__(f"Plugin '{name}' is not installed")


class PluginUpgradeError(Exception):
    """Base exception for plugin upgrade failures."""

    pass


class PluginVersionDowngradeError(PluginUpgradeError):
    """Attempted to upgrade to a version that is not newer."""

    def __init__(self, name: str, current_version: str, new_version: str):
        self.name = name
        self.current_version = current_version
        self.new_version = new_version
        super().__init__(
            f"Cannot upgrade plugin '{name}': new version {new_version} "
            f"is not greater than current version {current_version}"
        )



import os
from pathlib import Path

from hcli.lib.ida.plugin.repo import BasePluginRepo, Plugin, PluginArchiveIndex


class FileSystemPluginRepo(BasePluginRepo):
    def __init__(self, path: Path):
        super().__init__()
        self.path = path

    def get_plugins(self) -> list[Plugin]:
        index = PluginArchiveIndex()

        for root, dirs, files in os.walk(self.path):
            for file in files:
                if not file.endswith(".zip"):
                    continue

                path = Path(os.path.join(root, file))
                url = path.absolute().as_uri()

                with open(path, "rb") as f:
                    buf = f.read()

                index.index_plugin_archive(buf, url)

        return index.get_plugins()



from __future__ import annotations

import re

import questionary
import rich_click as click
from semantic_version import SimpleSpec

from hcli.commands.common import safe_ask_async
from hcli.env import ENV
from hcli.lib.commands import async_command
from hcli.lib.console import console
from hcli.lib.update.release import (
    GitHubRepo,
    get_assets,
    get_compatible_version,
    update_asset,
)
from hcli.lib.update.version import (
    is_binary,
)
from hcli.lib.util.io import get_arch, get_executable_path, get_os


@click.command()
@click.option("-f", "--force", is_flag=True, help="Force update.")
@click.option(
    "-m",
    "--mode",
    default="auto",
    type=click.Choice(["auto", "pypi", "binary"]),
    help="Update source (auto detects based on installation type).",
    hidden=True,
)
@click.option(
    "--auto-install", is_flag=True, help="Automatically install update if available (for binary version only)."
)
@click.option(
    "--include-prereleases",
    is_flag=True,
    help="Include pre-release versions when checking GitHub (for binary version only).",
)
@async_command
async def update(
    force: bool = False,
    mode: str = "auto",
    auto_install: bool = False,
    include_prereleases: bool = False,
) -> None:
    """Check for hcli updates."""

    # Auto-detect mode if not specified
    if mode == "auto":
        if is_binary():
            mode = "binary"  # Use GitHub for frozen binaries
        else:
            mode = "pypi"  # Use PyPI for non-frozen installs

    console.print(f"[bold]Checking for updates ({mode})...[/bold]")

    # Handle GitHub binary updates specially for frozen executables
    if mode == "binary" or is_binary():
        try:
            repo = GitHubRepo.from_url(ENV.HCLI_GITHUB_URL)

            # get current & latest
            current_version = ENV.HCLI_VERSION
            operator = ">=" if force else ">"
            latest_version = get_compatible_version(
                repo, SimpleSpec(f"{operator}{current_version}"), include_dev=include_prereleases
            )

            if latest_version is None:
                console.print(f"[green]Already using the latest version ({current_version})[/green]")
                return

            latest_tag = getattr(latest_version, "_origin_tag_name", None)
            assert latest_tag is not None, "Latest version tag not found"
            mask = f".*-{get_os()}-{get_arch()}.*"
            assets = get_assets(repo, latest_tag, re.compile(mask))

            if latest_tag and len(assets) == 1:
                console.print(f"[yellow]Update available: {current_version} → {latest_version}[/yellow]")

                # Skip confirmation if auto_install is enabled
                if not auto_install:
                    confirm = await safe_ask_async(
                        questionary.confirm(f"Do you want to install the update to {latest_version}?", default=True),
                        "Update cancelled.",
                    )
                    if not confirm:
                        console.print("[yellow]Update cancelled.[/yellow]")
                        return

                binary_path = get_executable_path()
                if not update_asset(repo, assets[0], binary_path):
                    console.print(f"[green]Already using the latest version ({current_version})[/green]")
                else:
                    console.print(f"[green]Successfully updated to {latest_version}[/green]")
                return

        except Exception as e:
            console.print(f"[red]Unexpected error during update: {e}[/red]")
            console.print("\nFalling back to manual update instructions...")
            console.print("\nTo update, run:")
            console.print("\nOn Mac or Linux, run:")
            console.print(f"[bold cyan]curl -LsSf {ENV.HCLI_RELEASE_URL}/install | sh[/bold cyan]")
            console.print("\nOr on Windows, run:")
            console.print(f"[bold cyan]iwr {ENV.HCLI_RELEASE_URL}/install.ps1 | iex[/bold cyan]")
            raise click.Abort()

    else:
        console.print("\nTo update, run:")
        console.print("[bold cyan]uv tool upgrade ida-hcli[/bold cyan]")
        console.print("or")
        console.print("[bold cyan]pipx upgrade ida-hcli[/bold cyan]")



from __future__ import annotations

from pathlib import Path

import questionary
import rich_click as click
from questionary import Choice
from rich.prompt import Confirm
from rich.table import Table

from hcli.commands.common import safe_ask_async
from hcli.lib.api.asset import SHARED, Asset, asset
from hcli.lib.api.common import get_api_client
from hcli.lib.commands import async_command, auth_command
from hcli.lib.console import console
from hcli.lib.constants import cli


@auth_command()
@click.option("--limit", type=int, default=100, help="Maximum number of files to display")
@click.option("--offset", type=int, default=0, help="Offset for pagination")
@click.option(
    "--interactive/--no-interactive",
    default=True,
    help="Enable interactive mode for file operations",
)
@async_command
async def list_shares(limit: int, offset: int, interactive: bool) -> None:
    """List and manage your shared files."""

    try:
        # Get shared files
        with console.status("[bold blue]Loading shared files..."):
            from hcli.lib.api.asset import PagingFilter

            filter_params = PagingFilter(limit=limit, offset=offset)
            page = await asset.get_files(SHARED, filter_params)

        if not page.items:
            console.print("[yellow]No shared files found.[/yellow]")
            return

        if interactive:
            await interactive_file_management(page.items)
        else:
            display_files_table(page.items)

    except Exception as e:
        console.print(f"[red]Error listing files: {e}[/red]")
        raise click.Abort()


def display_files_table(files: list[Asset]) -> None:
    """Display files in a formatted table."""
    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Index", style="dim", width=5)
    table.add_column("Code", style="cyan", width=10)
    table.add_column("Name", style="white", width=30)
    table.add_column("Version", style="blue", width=8)
    table.add_column("Size", style="green", width=10)
    table.add_column("Created", style="yellow", width=20)
    table.add_column("ACL", style="red", width=12)

    for i, file in enumerate(files, 1):
        name = truncate(file.filename or "unnamed", 30)
        version = f"v{file.version}"
        size = format_size(file.size)
        created = format_date_time(file.created_at) if file.created_at else "N/A"
        acl = file.metadata.get("acl_type") if file.metadata else None

        table.add_row(str(i), file.code, name, version, size, created, acl)

    console.print(table)


async def interactive_file_management(files: list[Asset]) -> None:
    """Provide interactive file management options."""

    # First, let user select files using checkbox
    file_choices = [
        Choice(title=f"{file.filename or 'unnamed'} ({file.code}) - {format_size(file.size)}", value=file)
        for file in files
    ]

    selected_files = await safe_ask_async(
        questionary.checkbox(
            "Select files to manage:",
            choices=file_choices,
            use_search_filter=True,
            use_jk_keys=False,
            style=cli.CHECKBOX_STYLE,
        )
    )

    if not selected_files:
        console.print("[yellow]No files selected.[/yellow]")
        return

    # Then let user choose action
    count = len(selected_files)
    action_choices = [
        f"Delete {count} file{'s' if count > 1 else ''}",
        f"Download {count} file{'s' if count > 1 else ''}",
    ]

    action = await safe_ask_async(
        questionary.select("What would you like to do?", choices=action_choices, style=cli.SELECT_STYLE)
    )

    if action.startswith("Delete"):
        await perform_delete_action(selected_files)
    elif action.startswith("Download"):
        await perform_download_action(selected_files)


async def perform_download_action(selected_files: list[Asset]) -> None:
    """Perform download action on selected files."""
    # Get output directory
    output_dir = await safe_ask_async(questionary.text("Output directory", default="./"))
    output_path = Path(output_dir)
    output_path.mkdir(parents=True, exist_ok=True)

    # Download files
    client = await get_api_client()

    for file in selected_files:
        try:
            console.print(f"\n[blue]Downloading {file.filename}...[/blue]")

            # Get download URL
            file_info = await asset.get_shared_file_by_code(file.code if file.code else "", file.version)

            if not file_info:
                console.print(f"[red]✗ Failed to get download info for {file.filename}[/red]")
                continue

            # Download file
            if not file_info.url:
                console.print(f"[red]✗ No download URL available for {file.filename}[/red]")
                continue

            downloaded_path = await client.download_file(
                file_info.url,
                target_dir=output_path,
                target_filename=file.filename,
                force=False,
                auth=True,
            )

            console.print(f"[green]✓ Downloaded: {downloaded_path}[/green]")

        except Exception as e:
            console.print(f"[red]✗ Failed to download {file.filename}: {e}[/red]")

    console.print(f"\n[green]Download completed. Files saved to: {output_path}[/green]")


async def perform_delete_action(selected_files: list[Asset]) -> None:
    """Perform delete action on selected files."""
    # Confirm deletion
    console.print(f"\n[red]You are about to delete {len(selected_files)} file(s):[/red]")
    for file in selected_files:
        console.print(f"  • {file.filename} ({file.code})")

    if not Confirm.ask("\n[bold red]Are you sure you want to delete these files?[/bold red]", default=False):
        console.print("[yellow]Deletion cancelled.[/yellow]")
        return

    # Delete files
    for file in selected_files:
        try:
            await asset.delete_file_by_key(SHARED, file.key)
            console.print(f"[green]✓ Deleted: {file.filename} ({file.code})[/green]")
        except Exception as e:
            console.print(f"[red]✗ Failed to delete {file.filename}: {e}[/red]")


def format_size(bytes_count: int) -> str:
    """Convert bytes to human-readable format."""
    sizes = ["B", "KB", "MB", "GB", "TB"]
    if bytes_count == 0:
        return "0 B"

    import math

    i = int(math.floor(math.log(bytes_count) / math.log(1024)))
    return f"{bytes_count / math.pow(1024, i):.1f} {sizes[i]}"


def format_date_time(iso_string: str) -> str:
    """Format ISO date string to readable format."""
    from datetime import datetime

    try:
        dt = datetime.fromisoformat(iso_string.replace("Z", "+00:00"))
        return dt.strftime("%Y-%m-%d %H:%M:%S")
    except Exception:
        return iso_string


def truncate(text: str, max_length: int) -> str:
    """Truncate text with ellipsis if too long."""
    return text[: max_length - 3] + "..." if len(text) > max_length else text



from __future__ import annotations

from pathlib import Path

import rich_click as click
from rich.console import Console
from rich.table import Table

from hcli.lib.config import config_store
from hcli.lib.ida import is_ida_dir

console = Console()


@click.command()
def list_instances() -> None:
    """List all registered IDA Pro instances."""
    # Get existing instances and default
    instances: dict[str, str] = config_store.get_object("ke.ida.instances", {}) or {}
    default_instance = config_store.get_string("ke.ida.default", "")

    if not instances:
        console.print("[yellow]No IDA Pro instances registered.[/yellow]")
        console.print("[yellow]Use 'hcli ke ida add --auto' to discover and add IDA installations.[/yellow]")
        return

    # Create table
    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Name", style="cyan", width=20)
    table.add_column("Path", style="white")
    table.add_column("Status", style="green", width=15)
    table.add_column("Default", style="yellow", width=8)

    # Add rows
    for name, path_str in instances.items():
        path = Path(path_str)

        # Check if the path still exists and is valid
        if path.exists() and is_ida_dir(path):
            status = "Valid"
            status_style = "green"
        elif path.exists():
            status = "Invalid"
            status_style = "red"
        else:
            status = "Missing"
            status_style = "red"

        # Check if this is the default
        is_default = "✓" if name == default_instance else ""

        table.add_row(name, str(path), f"[{status_style}]{status}[/{status_style}]", is_default)

    console.print(table)

    # Show summary
    valid_count = sum(1 for path_str in instances.values() if Path(path_str).exists() and is_ida_dir(Path(path_str)))
    total_count = len(instances)

    console.print(f"\n[blue]Summary:[/blue] {valid_count}/{total_count} instances are valid")

    if default_instance:
        if default_instance in instances:
            console.print(f"[blue]Default instance:[/blue] {default_instance}")
        else:
            console.print(f"[red]Default instance '{default_instance}' no longer exists![/red]")
    else:
        console.print("[yellow]No default instance set. Use 'hcli ke ida switch' to set one.[/yellow]")



from __future__ import annotations

from pathlib import Path

import rich_click as click

from hcli.lib.api.asset import asset
from hcli.lib.api.common import get_api_client
from hcli.lib.commands import async_command, auth_command
from hcli.lib.console import console


@auth_command()
@click.argument("shortcode", metavar="SHORTCODE")
@click.option(
    "-o",
    "--output-dir",
    "output_dir",
    type=click.Path(path_type=Path),
    help="Output directory (default: current directory)",
)
@click.option(
    "-O",
    "--output-file",
    "output_file",
    type=click.Path(path_type=Path),
    help="Output file path (conflicts with --output-dir)",
)
@click.option("-f", "--force", is_flag=True, help="Overwrite existing files")
@async_command
async def get(shortcode: str, output_dir: Path | None, output_file: Path | None, force: bool) -> None:
    """Download a shared file using its shortcode."""

    # Validate options
    if output_dir and output_file:
        console.print("[red]Error: --output-dir and --output-file cannot be used together[/red]")
        raise click.Abort()

    try:
        # Get file information
        with console.status("[bold blue]Getting file information..."):
            file_info = await asset.get_shared_file_by_code(shortcode)

        if not file_info:
            console.print(f"[red]Error: File with shortcode '{shortcode}' not found[/red]")
            return

        if output_file:
            output_file = output_file.expanduser().resolve()
            download_dir = output_file.parent
            filename = output_file.name
        else:
            download_dir = output_dir.expanduser().resolve() if output_dir else Path(".").resolve()
            filename = file_info.filename

        # Check if file already exists
        target_path = download_dir / filename
        if target_path.exists() and not force:
            console.print(f"[yellow]Warning: File already exists: {target_path}[/yellow]")
            if not click.confirm("Overwrite existing file?"):
                console.print("[yellow]Download cancelled[/yellow]")
                return

        # Download the file
        client = await get_api_client()
        if not file_info.url:
            console.print("[red]Error: No download URL available for file[/red]")
            return

        downloaded_path = await client.download_file(
            file_info.url,
            target_dir=download_dir,
            target_filename=filename,
            force=force,
            auth=True,
        )

        console.print("[green]✓ File downloaded successfully![/green]")
        console.print(f"[bold]File:[/bold] {file_info.filename}")
        console.print(f"[bold]Size:[/bold] {format_size(file_info.size)}")
        console.print(f"[bold]Saved to:[/bold] {downloaded_path}")

    except Exception as e:
        console.print(f"[red]Error downloading file: {e}[/red]")
        raise click.Abort()


def format_size(bytes_count: int) -> str:
    """Convert bytes to human-readable format."""
    sizes = ["B", "KB", "MB", "GB", "TB"]
    if bytes_count == 0:
        return "0 B"

    import math

    i = int(math.floor(math.log(bytes_count) / math.log(1024)))
    return f"{bytes_count / math.pow(1024, i):.1f} {sizes[i]}"



from __future__ import annotations

import rich_click as click


@click.group()
def source() -> None:
    """Manage knowledge sources."""
    pass


from .add import add  # noqa: E402
from .list import list_sources  # noqa: E402
from .remove import remove  # noqa: E402

source.add_command(add)
source.add_command(remove)
source.add_command(list_sources, name="list")



import json
from pathlib import Path
from typing import Any, TypeVar

from platformdirs import user_config_dir

from hcli.env import ENV

T = TypeVar("T")


class ConfigStore:
    """Cross-platform configuration storage for hcli."""

    def __init__(self):
        self._config_dir = Path(user_config_dir("hcli", "hex-rays"))
        self._config_file = self._config_dir / "config.json"
        self._data: dict[str, Any] = {}
        self._load_config()
        self._migrate_config()

    def _load_config(self) -> None:
        """Load configuration from disk."""
        if self._config_file.exists():
            try:
                with open(self._config_file, "r") as f:
                    self._data = json.load(f)
            except (json.JSONDecodeError, OSError):
                self._data = {}
        else:
            self._data = {}

    def _save_config(self):
        """Save configuration to disk."""
        self._config_dir.mkdir(parents=True, exist_ok=True)
        with open(self._config_file, "w") as f:
            json.dump(self._data, f, indent=2)

    def _migrate_config(self):
        """Migrate configuration if version changed."""
        current_version = self.get_string("version", "0.0.0")
        if current_version != ENV.HCLI_VERSION:
            self.set_string("version", ENV.HCLI_VERSION)
            self._save_config()

    def has(self, key: str) -> bool:
        """Check if key exists in configuration."""
        return key in self._data and self._data[key] is not None

    def get_string(self, key: str, default_value: str = "") -> str:
        """Get string value from configuration."""
        return self._data.get(key, default_value) or default_value

    def set_string(self, key: str, value: str = ""):
        """Set string value in configuration."""
        self._data[key] = value
        self._save_config()

    def remove_string(self, key: str):
        """Remove key from configuration."""
        if key in self._data:
            del self._data[key]
            self._save_config()

    def set_object(self, key: str, value: Any):
        """Set object value in configuration."""
        self._data[key] = value
        self._save_config()

    def get_object(self, key: str, default: T | None = None) -> T | None:
        """Get object value from configuration."""
        return self._data.get(key, default)


# Global config store instance
config_store = ConfigStore()



import functools
import json
import logging
import re
import time
import urllib.error
import urllib.parse
import urllib.request
from pathlib import Path
from typing import Any

import requests
import rich.progress
from pydantic import BaseModel, ConfigDict, Field
from tenacity import RetryCallState, retry, retry_if_exception, stop_after_attempt
from tenacity.wait import wait_base

from hcli.lib.console import stderr_console
from hcli.lib.ida.plugin.repo import BasePluginRepo, Plugin, PluginArchiveIndex
from hcli.lib.util.cache import get_cache_directory
from hcli.lib.util.logging import m

logger = logging.getLogger(__name__)

# Maximum file size to download (100MB)
MAX_DOWNLOAD_SIZE = 100 * 1024 * 1024

GITHUB_API_URL = "https://api.github.com"


def is_github_url(url: str) -> bool:
    """Check if URL is a GitHub repository URL."""
    return "github.com/" in url or url.startswith("git@github.com:")


def parse_github_url(url: str) -> tuple[str, str, str | None]:
    """Parse GitHub URL into (owner, repo, tag).

    Supports:
    - https://github.com/owner/repo
    - https://github.com/owner/repo.git
    - https://github.com/owner/repo@tag
    - git@github.com:owner/repo.git
    - git@github.com:owner/repo.git@tag
    """
    # Extract optional @tag suffix
    tag: str | None = None
    if "@" in url:
        # Handle git@ prefix separately
        if url.startswith("git@"):
            # git@github.com:owner/repo.git@tag
            parts = url.split(
                "@", 2
            )  # ['git', 'github.com:owner/repo.git', 'tag'] or ['git', 'github.com:owner/repo.git']
            if len(parts) == 3:
                tag = parts[2]
                url = f"git@{parts[1]}"
        else:
            # https://github.com/owner/repo@tag
            url, tag = url.rsplit("@", 1)

    # Parse the URL to get owner/repo
    if url.startswith("git@"):
        # git@github.com:owner/repo.git
        match = re.match(r"git@github\.com:([^/]+)/(.+?)(?:\.git)?$", url)
        if not match:
            raise ValueError(f"Invalid GitHub SSH URL: {url}")
        owner, repo = match.groups()
    else:
        # https://github.com/owner/repo(.git)
        parsed = urllib.parse.urlparse(url)
        path = parsed.path.lstrip("/")
        if path.endswith(".git"):
            path = path[:-4]
        parts = path.split("/")
        if len(parts) < 2:
            raise ValueError(f"Invalid GitHub URL: {url}")
        owner, repo = parts[0], parts[1]

    return owner, repo, tag


def fetch_github_release_zip_asset(owner: str, repo: str, tag: str | None = None) -> bytes:
    """Fetch .zip asset from GitHub release using REST API.

    Args:
        owner: Repository owner
        repo: Repository name
        tag: Release tag (e.g., "v1.0.0"). If None, fetches latest release.

    Returns:
        The .zip asset bytes.

    Raises:
        ValueError: If no .zip asset or multiple .zip assets found.
        requests.RequestException: If API request fails.
    """
    headers = {"Accept": "application/vnd.github.v3+json"}

    # Fetch release metadata
    if tag:
        release_url = f"{GITHUB_API_URL}/repos/{owner}/{repo}/releases/tags/{tag}"
    else:
        release_url = f"{GITHUB_API_URL}/repos/{owner}/{repo}/releases/latest"

    logger.info(f"fetching release from {release_url}")
    response = requests.get(release_url, headers=headers, timeout=30.0)
    response.raise_for_status()
    release_data = response.json()

    # Find .zip assets
    assets = release_data.get("assets", [])
    zip_assets = [a for a in assets if a.get("name", "").lower().endswith(".zip")]

    if not zip_assets:
        tag_info = f" ({tag})" if tag else " (latest)"
        raise ValueError(f"No .zip asset found in release{tag_info} for {owner}/{repo}")

    if len(zip_assets) > 1:
        asset_names = [a["name"] for a in zip_assets]
        raise ValueError(
            f"Multiple .zip assets found in release: {', '.join(asset_names)}. Cannot determine which to install."
        )

    asset = zip_assets[0]
    asset_name = asset["name"]
    asset_size = asset.get("size", 0)
    download_url = asset["browser_download_url"]

    if asset_size > MAX_DOWNLOAD_SIZE:
        raise ValueError(
            f"Asset {asset_name} ({asset_size} bytes) exceeds maximum size limit ({MAX_DOWNLOAD_SIZE} bytes)"
        )

    logger.info(f"downloading asset: {asset_name} ({asset_size} bytes) from {download_url}")
    response = requests.get(download_url, timeout=60.0)
    response.raise_for_status()

    return response.content


class WaitGitHubRateLimit(wait_base):
    """Custom wait strategy that respects GitHub's rate limit headers.

    GitHub provides these headers:
    - retry-after: seconds to wait before retrying
    - x-ratelimit-reset: UTC epoch timestamp when limit resets
    - x-ratelimit-remaining: requests remaining in current window

    If retry-after is present, use that.
    Otherwise, if x-ratelimit-remaining is 0, wait until x-ratelimit-reset.
    Otherwise, use exponential backoff with a minimum of 60 seconds.
    """

    def __init__(self, min_wait: int = 60, max_wait: int = 3600):
        self.min_wait = min_wait
        self.max_wait = max_wait

    def __call__(self, retry_state: RetryCallState) -> float:
        if retry_state.outcome and retry_state.outcome.failed:
            exception = retry_state.outcome.exception()
            if isinstance(exception, urllib.error.HTTPError):
                logger.debug(f"Rate limit headers received: {dict(exception.headers)}")

                retry_after = exception.headers.get("retry-after") or exception.headers.get("Retry-After")
                if retry_after:
                    retry_after_seconds = max(int(retry_after), self.min_wait)
                    logger.info(f"GitHub rate limit hit, respecting retry-after: {retry_after_seconds}s")
                    return min(retry_after_seconds, self.max_wait)

                remaining_str = exception.headers.get("x-ratelimit-remaining") or exception.headers.get(
                    "X-RateLimit-Remaining"
                )
                reset_time_str = exception.headers.get("x-ratelimit-reset") or exception.headers.get(
                    "X-RateLimit-Reset"
                )

                if reset_time_str:
                    reset_time = int(reset_time_str)
                    current_time = time.time()
                    raw_wait_time = reset_time - current_time

                    logger.debug(
                        f"Rate limit calculation: reset={reset_time}, current={current_time:.0f}, "
                        f"raw_wait={raw_wait_time:.0f}s, remaining={remaining_str}"
                    )

                    if raw_wait_time < 0:
                        logger.warning(
                            f"Clock skew detected: reset time {reset_time} is in the past "
                            f"(current time: {current_time:.0f}), using minimum wait"
                        )
                        wait_time = float(self.min_wait)
                    elif raw_wait_time > self.max_wait:
                        logger.warning(
                            f"Calculated wait time {raw_wait_time:.0f}s exceeds maximum {self.max_wait}s, "
                            f"capping to maximum"
                        )
                        wait_time = float(self.max_wait)
                    else:
                        wait_time = max(raw_wait_time, float(self.min_wait))

                    logger.info(f"GitHub rate limit exhausted, waiting until reset: {wait_time:.0f}s")
                    return wait_time

        attempt = retry_state.attempt_number
        exponential_wait = min(self.min_wait * (2 ** (attempt - 1)), self.max_wait)
        logger.info(f"GitHub rate limit hit, using exponential backoff: {exponential_wait:.0f}s (attempt {attempt})")
        return exponential_wait


def _is_rate_limit_error(exception: BaseException) -> bool:
    """Check if exception is a GitHub rate limit error (403 or 429)."""
    return isinstance(exception, urllib.error.HTTPError) and exception.code in (403, 429)


def _check_and_handle_proactive_rate_limit(response) -> None:
    """Check response headers and proactively wait if rate limit is nearly exhausted."""
    remaining_str = response.headers.get("x-ratelimit-remaining") or response.headers.get("X-RateLimit-Remaining")
    reset_time_str = response.headers.get("x-ratelimit-reset") or response.headers.get("X-RateLimit-Reset")

    if remaining_str and reset_time_str:
        remaining = int(remaining_str)
        if remaining <= 2:
            reset_time = int(reset_time_str)
            current_time = time.time()
            wait_time = max(reset_time - current_time, 30)

            if 0 < wait_time < 3600:
                logger.warning(
                    f"Proactive rate limit: only {remaining} requests remaining, "
                    f"waiting {wait_time:.0f}s until reset to avoid exhaustion"
                )
                time.sleep(wait_time)


@retry(
    retry=retry_if_exception(_is_rate_limit_error),
    wait=WaitGitHubRateLimit(min_wait=60, max_wait=3600),
    stop=stop_after_attempt(5),
    reraise=True,
)
def _urlopen_with_retry(req: urllib.request.Request):
    """Wrapper around urllib.request.urlopen with GitHub rate limit retry logic."""
    response = urllib.request.urlopen(req)
    _check_and_handle_proactive_rate_limit(response)
    return response


class GitHubReleaseAsset(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    name: str
    content_type: str = Field(alias="contentType")
    size: int
    download_url: str = Field(alias="downloadUrl")

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "GitHubReleaseAsset":
        return cls.model_validate(data)


class GitHubRelease(BaseModel):
    name: str
    tag_name: str
    commit_hash: str
    created_at: str
    published_at: str
    is_prerelease: bool
    is_draft: bool
    url: str
    zipball_url: str
    assets: list[GitHubReleaseAsset]

    @classmethod
    def from_dict(cls, data: dict[str, Any], owner: str, repo: str) -> "GitHubRelease":
        assets_data = data.get("releaseAssets", {}).get("nodes", [])
        assets = [GitHubReleaseAsset.from_dict(asset) for asset in assets_data]

        # Extract tarball and zipball URLs and commit hash from tag target
        tag_name = data.get("tagName", "")
        zipball_url = ""
        commit_hash = ""

        target = data["tag"]["target"]

        # release is against a tag
        # otherwise release is against a commit
        if "target" in target:
            target = target["target"]

        zipball_url = target["zipballUrl"]
        commit_hash = target["oid"]

        return cls(
            name=data.get("name", "") or data.get("tagName", ""),
            tag_name=tag_name,
            created_at=data["createdAt"],
            published_at=data["publishedAt"],
            is_prerelease=data["isPrerelease"],
            is_draft=data["isDraft"],
            url=data["url"],
            assets=assets,
            zipball_url=zipball_url,
            commit_hash=commit_hash,
        )


class GitHubTag(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    tag_name: str
    commit_hash: str
    zipball_url: str
    committed_date: str

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "GitHubTag":
        target = data["target"]
        if "target" in target:
            target = target["target"]

        return cls(
            tag_name=data["name"],
            commit_hash=target["oid"],
            zipball_url=target["zipballUrl"],
            committed_date=target["committedDate"],
        )


class GitHubCommit(BaseModel):
    commit_hash: str
    committed_date: str
    zipball_url: str

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "GitHubCommit":
        return cls(
            commit_hash=data["oid"],
            committed_date=data["committedDate"],
            zipball_url=data["zipballUrl"],
        )


class GitHubReleases(BaseModel):
    default_branch: GitHubCommit
    releases: list[GitHubRelease]
    tags: list[GitHubTag]


class GitHubGraphQLClient:
    """GitHub GraphQL API client"""

    def __init__(self, token: str):
        self.token = token
        self.api_url = "https://api.github.com/graphql"
        self.headers = {"Authorization": f"Bearer {self.token}", "Content-Type": "application/json"}

    def query(self, query: str, variables: dict[str, Any] | None = None) -> dict[str, Any]:
        """Execute a GraphQL query"""
        data = {"query": query, "variables": variables or {}}

        req = urllib.request.Request(self.api_url, data=json.dumps(data).encode("utf-8"), headers=self.headers)

        try:
            with _urlopen_with_retry(req) as response:
                result = json.loads(response.read().decode("utf-8"))

                if "errors" in result:
                    raise Exception(f"GraphQL errors: {result['errors']}")

                return result["data"]
        except urllib.error.HTTPError as e:
            error_body = e.read().decode("utf-8")
            raise Exception(f"HTTP {e.code}: {error_body}")

    def get_many_releases(self, repos: list[tuple[str, str]], count: int = 10) -> dict[tuple[str, str], GitHubReleases]:
        """Fetch releases for multiple repositories in a single query

        Returns: mapping from (owner, repo) -> GitHubReleases
        """
        if not repos:
            return {}

        logger.info(f"fetching releases from GitHub API for {len(repos)} repositories")

        # Build query with aliases
        query_parts = []
        variables = {"first": count}

        for i, (owner, repo) in enumerate(repos):
            alias = f"repo{i}"
            query_parts.append(f"""
                {alias}: repository(owner: "{owner}", name: "{repo}") {{
                    defaultBranchRef {{
                        target {{
                            ... on Commit {{
                                oid
                                zipballUrl
                                committedDate
                            }}
                        }}
                    }}
                    releases(first: $first, orderBy: {{field: CREATED_AT, direction: DESC}}) {{
                        nodes {{
                            name
                            tagName
                            createdAt
                            publishedAt
                            isPrerelease
                            isDraft
                            url
                            releaseAssets(first: 50) {{
                                nodes {{
                                    name
                                    downloadUrl
                                    size
                                    contentType
                                }}
                            }}
                            tag {{
                                target {{
                                    ... on Commit {{
                                        zipballUrl
                                        oid
                                        committedDate
                                    }}
                                    ... on Tag {{
                                        target {{
                                            ... on Commit {{
                                                zipballUrl
                                                oid
                                                committedDate
                                            }}
                                        }}
                                    }}
                                }}
                            }}
                        }}
                    }}
                    refs(refPrefix: "refs/tags/", first: 25, orderBy: {{field: TAG_COMMIT_DATE, direction: DESC}}) {{
                        nodes {{
                            name
                            target {{
                                ... on Commit {{
                                    zipballUrl
                                    oid
                                    committedDate
                                }}
                                ... on Tag {{
                                    target {{
                                        ... on Commit {{
                                            zipballUrl
                                            oid
                                            committedDate
                                        }}
                                    }}
                                }}
                            }}
                        }}
                    }}
                }}
            """)

        query = f"""
        query($first: Int!) {{
            {"".join(query_parts)}
        }}
        """

        data = self.query(query, variables)

        result = {}
        for i, (owner, repo) in enumerate(repos):
            repo_data = data.get(f"repo{i}")

            if not repo_data:
                logger.warning(f"Repository {owner}/{repo} not found")
                continue

            releases_data = repo_data["releases"]["nodes"]
            tags_data = repo_data["refs"]["nodes"]
            result[(owner, repo)] = GitHubReleases(
                default_branch=GitHubCommit.from_dict(repo_data["defaultBranchRef"]["target"]),
                releases=[GitHubRelease.from_dict(release_data, owner, repo) for release_data in releases_data],
                tags=[GitHubTag.from_dict(tag_data) for tag_data in tags_data],
            )

        return result

    def get_releases(self, owner: str, repo: str, count: int = 10) -> GitHubReleases:
        key = (owner, repo)
        return self.get_many_releases([key])[key]


def parse_repository(repo_string: str) -> tuple[str, str]:
    """Parse repository string into owner and repo name"""
    if "/" not in repo_string:
        raise ValueError(f"invalid repository format: {repo_string}. Expected format: owner/repo")

    parts = repo_string.split("/")
    if len(parts) != 2:
        raise ValueError(f"invalid repository format: {repo_string}. Expected format: owner/repo")

    return parts[0], parts[1]


def get_source_archive_cache_directory(owner: str, repo: str, commit_hash: str) -> Path:
    return get_cache_directory(owner, repo, "source-archives", commit_hash)


def get_release_asset_cache_directory(owner: str, repo: str, release_id: str) -> Path:
    return get_cache_directory(owner, repo, "release-assets", release_id)


def get_releases_metadata_cache_path(owner: str, repo: str) -> Path:
    return get_cache_directory(owner, repo) / "releases.json"


def set_releases_metadata_cache(owner: str, repo: str, releases: GitHubReleases) -> None:
    cache_path = get_releases_metadata_cache_path(owner, repo)
    releases_data = releases.model_dump()
    cache_path.write_text(json.dumps(releases_data, indent=2, sort_keys=True))
    logger.debug(f"saved releases cache to: {cache_path}")


def get_releases_metadata_cache(owner: str, repo: str) -> GitHubReleases:
    cache_path = get_releases_metadata_cache_path(owner, repo)
    if not cache_path.exists():
        raise KeyError(f"no releases cache found for {owner}/{repo}")

    file_age = time.time() - cache_path.stat().st_mtime

    # release metadata cache expires after 24 hours
    # based on file modification time
    if file_age > 24 * 60 * 60:  # 24 hours
        logger.info(f"cache expired for {owner}/{repo} releases metadata, removing file")
        cache_path.unlink()
        raise KeyError(f"expired releases cache removed for {owner}/{repo}")

    releases_data = json.loads(cache_path.read_text())
    return GitHubReleases.model_validate(releases_data)


def warm_releases_metadata_cache(client: GitHubGraphQLClient, repos: list[tuple[str, str]]) -> None:
    """Warm the releases metadata cache for multiple repositories"""

    repos_to_fetch = []

    for owner, repo in repos:
        try:
            get_releases_metadata_cache(owner, repo)
        except KeyError:
            repos_to_fetch.append((owner, repo))

    if not repos_to_fetch:
        logger.debug("all repositories already cached")
        return

    logger.debug(f"warming cache for {len(repos_to_fetch)} repositories")

    BATCH_SIZE = 10
    for i in rich.progress.track(
        range(0, len(repos_to_fetch), BATCH_SIZE), description="Warming cache", transient=True, console=stderr_console
    ):
        batch = repos_to_fetch[i : i + BATCH_SIZE]
        releases_batch = client.get_many_releases(batch)

        for (owner, repo), releases in releases_batch.items():
            set_releases_metadata_cache(owner, repo, releases)


def get_releases_metadata(client: GitHubGraphQLClient, owner: str, repo: str) -> GitHubReleases:
    try:
        return get_releases_metadata_cache(owner, repo)
    except KeyError:
        releases = client.get_releases(owner, repo)
        set_releases_metadata_cache(owner, repo, releases)
        return releases


def set_release_asset_cache(owner: str, repo: str, release_id: str, asset: GitHubReleaseAsset, buf: bytes):
    cache_path = get_release_asset_cache_directory(owner, repo, release_id)
    (cache_path / asset.name).write_bytes(buf)
    logger.debug(f"asset {asset.name} cached for {owner}/{repo} release {release_id}")


def get_release_asset_cache(owner: str, repo: str, release_id: str, asset: GitHubReleaseAsset) -> bytes:
    cache_path = get_release_asset_cache_directory(owner, repo, release_id)
    asset_path = cache_path / asset.name
    if not asset_path.exists():
        raise KeyError(f"asset {asset.name} not found in cache for {owner}/{repo} release {release_id}")

    logger.debug(f"asset {asset.name} found in cache for {owner}/{repo} release {release_id}")
    return asset_path.read_bytes()


def download_release_asset(owner: str, repo: str, release_id: str, asset: GitHubReleaseAsset) -> bytes:
    if asset.size > MAX_DOWNLOAD_SIZE:
        raise ValueError(f"asset {asset.name} exceeds {MAX_DOWNLOAD_SIZE} limit")

    logger.info(f"downloading asset: {asset.name} ({asset.size}) from {asset.download_url}")
    req = urllib.request.Request(asset.download_url)
    with _urlopen_with_retry(req) as response:
        asset_data = response.read()

    logger.debug(f"downloaded {len(asset_data)} bytes for asset {asset.name}")
    return asset_data


def get_release_asset(owner: str, repo: str, release_id: str, asset: GitHubReleaseAsset) -> bytes:
    try:
        return get_release_asset_cache(owner, repo, release_id, asset)
    except KeyError:
        buf = download_release_asset(owner, repo, release_id, asset)
        set_release_asset_cache(owner, repo, release_id, asset, buf)
        return buf


SOURCE_ARCHIVE_FILENAME = "source.zip"


def set_source_archive_cache(owner: str, repo: str, commit_hash: str, buf: bytes):
    cache_path = get_source_archive_cache_directory(owner, repo, commit_hash)
    (cache_path / SOURCE_ARCHIVE_FILENAME).write_bytes(buf)
    logger.debug(f"Source archive cached for {owner}/{repo}@{commit_hash[:8]}")


def get_source_archive_cache(owner: str, repo: str, commit_hash: str) -> bytes:
    cache_path = get_source_archive_cache_directory(owner, repo, commit_hash)
    archive_path = cache_path / SOURCE_ARCHIVE_FILENAME
    if not archive_path.exists():
        raise KeyError(f"source archive not found in cache for {owner}/{repo}@{commit_hash[:8]}")

    logger.debug(f"source archive found in cache for {owner}/{repo}@{commit_hash[:8]}")
    return archive_path.read_bytes()


def download_source_archive(zip_url: str) -> bytes:
    logger.info(f"downloading source archive from {zip_url}")
    req = urllib.request.Request(zip_url)
    with _urlopen_with_retry(req) as response:
        buf = response.read()

    logger.debug(f"downloaded {len(buf)} bytes from {zip_url}")
    return buf


def get_source_archive(owner: str, repo: str, commit_hash: str, zip_url: str) -> bytes:
    try:
        return get_source_archive_cache(owner, repo, commit_hash)
    except KeyError:
        buf = download_source_archive(zip_url)
        set_source_archive_cache(owner, repo, commit_hash, buf)
        return buf


def get_release_metadata(client: GitHubGraphQLClient, owner: str, repo: str, release_id: str) -> GitHubRelease:
    """Extract release metadata from a release"""
    for release in get_releases_metadata(client, owner, repo).releases:
        if release.tag_name == release_id:
            return release

    raise KeyError(f"release {release_id} not found for {owner}/{repo}")


def get_candidate_github_repos_cache_path() -> Path:
    return get_cache_directory() / "candidate_repos.json"


def set_candidate_github_repos_cache(repos: list[str]) -> None:
    cache_path = get_candidate_github_repos_cache_path()
    cache_path.write_text(json.dumps(repos, indent=2, sort_keys=True))
    logger.debug(f"Saved candidate repos cache to: {cache_path}")


def get_candidate_github_repos_cache() -> list[str]:
    cache_path = get_candidate_github_repos_cache_path()
    if not cache_path.exists():
        raise KeyError("no candidate repos cache found")

    file_age = time.time() - cache_path.stat().st_mtime

    # release metadata cache expires after 24 hours
    # based on file modification time
    if file_age > 24 * 60 * 60:  # 24 hours
        logger.info("cache expired for candidate repos, removing file")
        cache_path.unlink()
        raise KeyError("expired candidate repos cache")

    return json.loads(cache_path.read_text())


def find_github_repos_with_plugins(token: str) -> list[str]:
    """Find GitHub repositories that contain ida-plugin.json files using GitHub's search API.

    Returns:
        List of repositories in "owner/repo" format
    """
    BATCH_SIZE = 25

    # Note: Forks with fewer stars than the parent repository or no commits are not indexed for code search.
    # via: https://docs.github.com/en/search-github/searching-on-github/searching-code
    queries = [
        "filename:ida-plugin.json",
        "filename:ida-plugin.json fork:true",
    ]

    repos = set()
    for query in queries:
        search_url = "https://api.github.com/search/code"
        headers = {
            "Authorization": f"Bearer {token}",
            "Accept": "application/vnd.github.v3+json",
            "User-Agent": "ida-hcli",
        }

        page = 1
        while True:
            params = f"q={urllib.parse.quote(query)}&per_page={BATCH_SIZE}&page={page}"
            url = f"{search_url}?{params}"

            req = urllib.request.Request(url, headers=headers)
            with _urlopen_with_retry(req) as response:
                result = json.loads(response.read().decode("utf-8"))

                items = result.get("items", [])
                if not items:
                    break

                for item in items:
                    repo_full_name = item["repository"]["full_name"]
                    repos.add(repo_full_name)
                    logger.debug(
                        m(
                            "found repository via GitHub search: %s",
                            repo_full_name,
                            query=query,
                            page=page,
                            name=repo_full_name,
                        )
                    )

                if len(items) < BATCH_SIZE:
                    break

                page += 1

    return sorted(list(repos))


class GithubPluginRepo(BasePluginRepo):
    def __init__(self, token: str, extra_repos: list[str] | None = None, ignored_repos: list[str] | None = None):
        super().__init__()
        self.token = token
        self.extra_repos = set(extra_repos or [])
        self.ignored_repos = set(ignored_repos or [])
        self.client = GitHubGraphQLClient(token)

        # warm cache
        self._repos = self._get_repos()

        warm_releases_metadata_cache(self.client, self._repos)

    def _get_repos(self):
        try:
            repos = set(get_candidate_github_repos_cache())
        except KeyError:
            repos = set(find_github_repos_with_plugins(self.token))
            repos = {r.lower() for r in repos}
            set_candidate_github_repos_cache(sorted(repos))
        else:
            repos = {r.lower() for r in repos}

        extra = {r.lower() for r in self.extra_repos}
        ignored = {r.lower() for r in self.ignored_repos}

        for repo in sorted(extra & repos):
            logger.debug("extra repo already found by GitHub index: %s", repo)
        for repo in sorted(extra - repos):
            logger.debug("extra repo not yet found by GitHub index: %s", repo)
        for repo in sorted(repos - extra):
            logger.debug("GitHub repo not in extra repo list: %s", repo)
        repos |= extra

        for repo in sorted(ignored & repos):
            logger.debug("ignoring found repo: %s", repo)
        repos -= ignored

        return [parse_repository(repo) for repo in sorted(repos)]

    @functools.cache
    def get_plugins(self) -> list[Plugin]:
        assets = []
        source_archives = []

        # first collect all the URLs
        # then fetch them in a second loop
        # so that we can have a meaningful progress bar.

        for owner, repo in sorted(self._repos):
            logger.debug("finding plugins in repo: %s/%s", owner, repo)

            md = get_releases_metadata(self.client, owner, repo)
            seen_zipball_urls = set()
            for release in md.releases:
                context = {
                    "owner": owner,
                    "repo": repo,
                    "release": release.tag_name,
                    "date": release.published_at,
                }
                logger.debug(m("considering release: %s", release.tag_name, **context))

                if release.published_at < "2025-09-01":
                    logger.debug(
                        m(
                            "skipping old release: %s < 2025-09-01",
                            release.published_at,
                            **context,
                        )
                    )
                    continue

                logger.debug(m("found release: %s", release.tag_name, **context))

                # source archives
                source_archives.append((owner, repo, release.commit_hash, release.zipball_url, release.published_at))
                seen_zipball_urls.add(release.zipball_url)
                logger.debug(m("found zipball URL: %s", release.zipball_url, **context))

                # assets (distribution/binary archives)
                for asset in release.assets:
                    if asset.content_type not in {"application/zip", "application/x-zip-compressed", "raw"}:
                        # GitHub provides various mimetypes for manually attached ZIP archives, including:
                        # - https://github.com/binsync/binsync/releases/download/v5.10.1/binsync-ida-plugin.zip (raw)
                        # - https://github.com/arkup/tc_deer/releases/download/0.1.0/tc_deer_ida_plugin010.zip (application/x-zip-compressed)
                        # so we loosen the content type restrictions, but enforce filename ending with .zip below
                        logger.debug(
                            m(
                                "skipping asset with type: %s",
                                asset.content_type,
                                **dict(context, asset=asset.name),
                            )
                        )
                        continue

                    if not asset.name.lower().endswith(".zip"):
                        logger.debug(
                            m(
                                "skipping asset with name: %s",
                                asset.name,
                                **dict(context, asset=asset.name),
                            )
                        )
                        continue

                    assets.append((owner, repo, release.tag_name, asset, release.published_at))
                    logger.debug(
                        m(
                            "found zip asset: %s",
                            asset.download_url,
                            **dict(context, asset=asset.name),
                        )
                    )

            for tag in md.tags:
                context = {
                    "owner": owner,
                    "repo": repo,
                    "tag": tag.tag_name,
                    "date": tag.committed_date,
                }

                logger.debug(m("considering tag: %s", tag.tag_name, **context))

                if not tag.tag_name.startswith("v"):
                    logger.debug(m("skipping non-v* tag: %s", tag.tag_name, **context))
                    continue

                if tag.committed_date < "2025-09-01":
                    logger.debug(
                        m(
                            "skipping old tag: %s < 2025-09-01",
                            tag.committed_date,
                            **context,
                        )
                    )
                    continue

                logger.debug(m("found tag: %s", tag.tag_name, **context))

                if tag.zipball_url in seen_zipball_urls:
                    logger.debug(m("already found URL for tag: %s", tag.zipball_url, **context))
                else:
                    source_archives.append((owner, repo, tag.commit_hash, tag.zipball_url, tag.committed_date))
                    seen_zipball_urls.add(tag.zipball_url)
                    logger.debug(m("found zipball URL: %s", tag.zipball_url, **context))

        index = PluginArchiveIndex()

        for owner, repo, tag_name, asset, date in rich.progress.track(
            assets, description="Fetching plugin assests", transient=True, console=stderr_console
        ):
            logger.debug(m("fetching release asset: %s", asset.download_url, owner=owner, repo=repo, tag=tag_name))
            try:
                buf = get_release_asset(owner, repo, tag_name, asset)
            except ValueError:
                continue

            host_url = f"https://github.com/{owner}/{repo}"
            index.index_plugin_archive(
                buf,
                asset.download_url,
                expected_host=host_url,
                context=dict(
                    owner=owner, repo=repo, type="release asset", tag=tag_name, url=asset.download_url, date=date
                ),
            )

        for owner, repo, commit_hash, url, date in rich.progress.track(
            source_archives, description="Fetching plugin source archives", transient=True, console=stderr_console
        ):
            logger.debug(m("fetching source archive: %s", url, owner=owner, repo=repo, commit=commit_hash))
            try:
                buf = get_source_archive(owner, repo, commit_hash, url)
            except ValueError:
                continue

            host_url = f"https://github.com/{owner}/{repo}"
            index.index_plugin_archive(
                buf,
                url,
                expected_host=host_url,
                context=dict(owner=owner, repo=repo, type="source archive", commit=commit_hash, url=url, date=date),
            )

        return index.get_plugins()



"""repository management commands."""

from __future__ import annotations

import logging

import rich_click as click

from hcli.lib.console import console
from hcli.lib.ida.plugin.repo.file import JSONFilePluginRepo

logger = logging.getLogger(__name__)


@click.group(hidden=True)
@click.pass_context
def repo(ctx) -> None:
    """Manage plugin repositories."""
    pass


@repo.command()
@click.pass_context
def snapshot(ctx) -> None:
    """Create a snapshot of the repository."""
    try:
        repo = JSONFilePluginRepo.from_repo(ctx.obj["plugin_repo"])
        # Use print() instead of console.print() to output raw JSON without ANSI control characters
        print(repo.to_json())
    except Exception as e:
        logger.debug("error: %s", e, exc_info=True)
        console.print(f"[red]Error[/red]: {e}")
        raise click.Abort()



from __future__ import annotations

import logging
from pathlib import Path

import requests
import rich_click as click

from hcli.lib.console import console
from hcli.lib.ida import (
    FailedToDetectIDAVersion,
    MissingCurrentInstallationDirectory,
    explain_failed_to_detect_ida_version,
    explain_missing_current_installation_directory,
    find_current_ida_platform,
    find_current_ida_version,
)
from hcli.lib.ida.plugin import get_metadata_from_plugin_archive
from hcli.lib.ida.plugin.install import upgrade_plugin_archive
from hcli.lib.ida.plugin.repo import BasePluginRepo

logger = logging.getLogger(__name__)


@click.command()
@click.pass_context
@click.argument("plugin")
def upgrade_plugin(ctx, plugin: str) -> None:
    """Upgrade an installed plugin to the latest compatible version."""
    plugin_spec = plugin
    try:
        current_ida_platform = find_current_ida_platform()
        current_ida_version = find_current_ida_version()

        if Path(plugin_spec).exists() and plugin_spec.endswith(".zip"):
            raise ValueError("cannot upgrade using local file; uninstall/reinstall instead")

        if plugin_spec.startswith("file://"):
            raise ValueError("cannot upgrade using local file; uninstall/reinstall instead")

        if plugin_spec.startswith("https://"):
            raise ValueError("cannot upgrade using URL; uninstall/reinstall instead")

        logger.info("finding plugin in repository")
        plugin_repo: BasePluginRepo = ctx.obj["plugin_repo"]
        try:
            plugin_name, buf = plugin_repo.fetch_compatible_plugin_from_spec(
                plugin_spec, current_ida_platform, current_ida_version
            )
        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
            console.print("[red]Cannot connect to plugin repository - network unavailable.[/red]")
            console.print("Please check your internet connection.")
            raise click.Abort()

        upgrade_plugin_archive(buf, plugin_name)

        _, metadata = get_metadata_from_plugin_archive(buf, plugin_name)

        console.print(f"[green]Installed[/green] plugin: [blue]{plugin_name}[/blue]=={metadata.plugin.version}")
    except MissingCurrentInstallationDirectory:
        explain_missing_current_installation_directory(console)
        raise click.Abort()

    except FailedToDetectIDAVersion:
        explain_failed_to_detect_ida_version(console)
        raise click.Abort()

    except Exception as e:
        logger.debug("error: %s", e, exc_info=True)
        console.print(f"[red]Error[/red]: {e}")
        raise click.Abort()



# see also hcli.lib.util.python
import logging
import os
import subprocess
from pathlib import Path

from hcli.env import ENV
from hcli.lib.ida import run_py_in_current_idapython

logger = logging.getLogger(__name__)


FIND_PYTHON_PY = """
# output like:
#
#     __hcli__:"/Users/user/.idapro/venv/bin/python"
import shutil
import os.path
import sys
import json
def find_python_executable():
    exe = sys.executable
    if "python" in os.path.basename(exe).lower():
      return exe

    return shutil.which("python3") or shutil.which("python")
print("__hcli__:" + json.dumps(find_python_executable()))
sys.exit()
"""


def find_current_python_executable() -> Path:
    """find the python executable associated with the current IDA installation"""
    # duplicate here, because we prefer access through ENV
    # but tests might update env vars for the current process.
    exe = os.environ.get("HCLI_CURRENT_IDA_PYTHON_EXE")
    if exe:
        return Path(exe)
    if ENV.HCLI_CURRENT_IDA_PYTHON_EXE is not None:
        return Path(ENV.HCLI_CURRENT_IDA_PYTHON_EXE)

    try:
        exe = run_py_in_current_idapython(FIND_PYTHON_PY)
    except RuntimeError as e:
        raise RuntimeError("failed to determine current IDA Python interpreter") from e

    if "python" not in exe:
        logger.warning("'python' not found in discovered IDA Python interpreter path: %s", exe)

    logger.debug("found IDA Python interpreter path: %s", exe)
    return Path(exe)


def does_current_ida_have_pip(python_exe: Path) -> bool:
    """Check if pip is available in the given Python executable."""
    try:
        process = subprocess.run(
            [str(python_exe), "-m", "pip", "help"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=10.0
        )
        return process.returncode == 0
    except (subprocess.TimeoutExpired, FileNotFoundError, OSError):
        return False


class CantInstallPackagesError(ValueError): ...


def verify_pip_can_install_packages(python_exe: Path, packages: list[str]):
    """Check if the given Python packages (e.g., "foo>=v1.0,<3") can be installed.

    This allows pip to determine if there are any version conflicts
    """
    process = subprocess.run(
        [str(python_exe), "-m", "pip", "install", "--dry-run"] + packages,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = process.stdout, process.stderr
    if process.returncode != 0:
        # error output might look like:
        #
        #     ❯ pip install --dry-run flare-capa==v1.0.0 flare-capa==v1.0.1
        #    Collecting flare-capa==v1.0.0
        #      Using cached flare-capa-1.0.0.tar.gz (62 kB)
        #      Installing build dependencies ... done
        #      Getting requirements to build wheel ... done
        #      Preparing metadata (pyproject.toml) ... done
        #    ERROR: Cannot install flare-capa==v1.0.0 and flare-capa==v1.0.1 because these package versions have conflicting dependencies.
        #
        #    The conflict is caused by:
        #        The user requested flare-capa==v1.0.0
        #        The user requested flare-capa==v1.0.1
        #
        #    To fix this you could try to:
        #    1. loosen the range of package versions you've specified
        #    2. remove package versions to allow pip to attempt to solve the dependency conflict
        #
        #    ERROR: ResolutionImpossible: for help visit https://pip.pypa.io/en/latest/topics/dependency-resolution/#dealing-with-dependency-conflicts
        logger.debug("can't install packages")
        logger.debug(stdout.decode())
        logger.debug(stderr.decode())
        raise CantInstallPackagesError(stdout.decode())


def pip_install_packages(python_exe: Path, packages: list[str]):
    """Install the given Python packages (e.g., "foo>=v1.0,<3")."""
    process = subprocess.run(
        [str(python_exe), "-m", "pip", "install"] + packages, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    stdout, stderr = process.stdout, process.stderr
    if process.returncode != 0:
        # error output might look like:
        #
        #     ❯ pip install --dry-run flare-capa==v1.0.0 flare-capa==v1.0.1
        #    Collecting flare-capa==v1.0.0
        #      Using cached flare-capa-1.0.0.tar.gz (62 kB)
        #      Installing build dependencies ... done
        #      Getting requirements to build wheel ... done
        #      Preparing metadata (pyproject.toml) ... done
        #    ERROR: Cannot install flare-capa==v1.0.0 and flare-capa==v1.0.1 because these package versions have conflicting dependencies.
        #
        #    The conflict is caused by:
        #        The user requested flare-capa==v1.0.0
        #        The user requested flare-capa==v1.0.1
        #
        #    To fix this you could try to:
        #    1. loosen the range of package versions you've specified
        #    2. remove package versions to allow pip to attempt to solve the dependency conflict
        #
        #    ERROR: ResolutionImpossible: for help visit https://pip.pypa.io/en/latest/topics/dependency-resolution/#dealing-with-dependency-conflicts
        logger.debug("can't install packages")
        logger.debug(stdout.decode())
        logger.debug(stderr.decode())
        raise CantInstallPackagesError(stdout.decode())


def pip_freeze(python_exe: Path):
    process = subprocess.run([str(python_exe), "-m", "pip", "freeze"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, _ = process.stdout, process.stderr
    if process.returncode != 0:
        raise subprocess.CalledProcessError(process.returncode, [str(python_exe), "-m", "pip", "freeze"])
    return stdout.decode()



import rich_click as click
from rich.console import Console


def __get_console() -> Console:
    """Get console instance with quiet mode support."""
    try:
        ctx = click.get_current_context(silent=True)
        if ctx and ctx.obj and ctx.obj.get("quiet", False):
            return Console(quiet=True)
    except RuntimeError:
        # No context available, return default console
        pass
    return Console()


def __get_stderr_console() -> Console:
    """Get console instance with quiet mode support."""
    try:
        ctx = click.get_current_context(silent=True)
        if ctx and ctx.obj and ctx.obj.get("quiet", False):
            return Console(quiet=True, stderr=True)
    except RuntimeError:
        # No context available, return default console
        pass
    return Console(stderr=True)


# Global instances for convenience
console = __get_console()
stderr_console = __get_stderr_console()



from __future__ import annotations

import json
from pathlib import Path

import rich_click as click
from rich.prompt import Confirm

from hcli.commands.download import download
from hcli.commands.license.get import get_license
from hcli.lib.auth import get_auth_service
from hcli.lib.commands import async_command, enforce_login
from hcli.lib.console import console
from hcli.lib.ida import (
    IdaProduct,
    accept_eula,
    get_default_ida_install_directory,
    get_ida_config_path,
    get_ida_path,
    get_license_dir,
    install_ida,
    install_license,
)
from hcli.lib.util.io import get_os, get_temp_dir


@click.option(
    "-d",
    "--download-id",
    "download_id",
    required=False,
    help="Full installer asset key, or tag (e.g., 'ida-pro:latest', 'ida-essential:9.2')",
)
@click.option("-l", "--license-id", "license_id", required=False, help="License ID (e.g., 96-0000-0000-01)")
@click.option("-i", "--install-dir", "install_dir", required=False, help="Install dir")
@click.option("-a", "--accept-eula", "eula", is_flag=True, help="Accept EULA", default=True)
@click.option("--set-default", is_flag=True, help="Mark this IDA installation as the default", default=False)
@click.option("--dry-run", is_flag=True, help="Show what would be done without actually installing")
@click.option("--yes", "-y", "auto_confirm", is_flag=True, help="Auto-accept confirmation prompts", default=False)
@click.argument("installer", required=False)
@click.command()
@click.pass_context
@async_command
async def install(
    ctx,
    install_dir: str | None,
    eula: bool,
    installer: str,
    download_id: str | None,
    license_id: str | None,
    set_default: bool,
    dry_run: bool,
    auto_confirm: bool,
) -> None:
    """Installs IDA unattended.

    \b
    The --download-id option supports tags for simplified version specification:
    - 'category:version' (e.g., 'ida-pro:latest') - OS is auto-detected
    - 'category:version:os' (e.g., 'ida-pro:9.2:x64linux') - explicit OS

    \b
    If install_dir is /tmp/myida, the ida binary will be located:
    - on Windows: /tmp/myida/ida
    - on Linux: /tmp/myida/ida
    - on Mac: /tmp/myida/Contents/MacOS/ida
    """
    try:
        tmp_dir = get_temp_dir()

        if download_id or license_id:
            auth_service = get_auth_service()
            auth_service.init()

            enforce_login()

        # download installer using the download command
        if download_id:
            # Download the installer (may be a tag like 'ida-pro:latest' or a full key)
            await download.callback(output_dir=tmp_dir, key=download_id)

            # Find the downloaded installer file in tmp_dir
            # Look for common IDA installer extensions
            installer_files: list[Path] = []
            for pattern in ["*.app.zip", "*.run", "*.exe", "*.zip"]:
                installer_files.extend(Path(tmp_dir).glob(pattern))

            if not installer_files:
                raise FileNotFoundError(f"No installer file found in {tmp_dir} after download")

            # Use the most recently modified file (should be the one we just downloaded)
            installer_path = max(installer_files, key=lambda p: p.stat().st_mtime).resolve()
        elif installer is not None:
            installer_path = Path(installer).resolve()
        else:
            raise click.UsageError("Either provide an installer file path or use --download-id to download one")

        if not installer_path.exists():
            raise FileNotFoundError(
                f"Installer file not found: {installer_path}\nPlease ensure the file exists at the specified location."
            )

        version = IdaProduct.from_installer_filename(installer_path.name)

        if not install_dir:
            install_dir_path = get_default_ida_install_directory(version)
        else:
            install_dir_path = Path(install_dir).expanduser().resolve()

        # prominent warning for #99: idat from IDA 9.2 on Linux fails to start if the path contains a space.
        #
        # typically idat isn't widely used; however, HCLI does use it to discover the path to IDA's Python interpreter,
        # as well as the installed arch (ARM or Intel on macOS). The latter could probably be discovered by inspecting
        # the installed files; however, figuring out the Python configuration is messy, and much easier to leave to idat.
        #
        # see also the special handling in hcli.lib.ida
        if get_os() == "linux" and version.major == 9 and version.minor == 2 and install_dir:
            if " " in str(install_dir_path.absolute()):
                console.print(
                    "[yellow]Warning[/yellow]: Avoid installation paths with a space for this release (IDA 9.2 on Linux).\n",
                    "This specific release of IDA has a bug in idat, preventing it from working\n",
                    "when the path contains a space.\n\n",
                    "You can find alternative workarounds in #99.\n\n",
                    f"[grey69]Current directory name:[/grey69] '{install_dir_path}'",
                )

                if not auto_confirm:
                    if not Confirm.ask("[bold yellow]Continue anyway?[/bold yellow]", default=False):
                        console.print("[yellow]Installation cancelled.[/yellow]")
                        return

        console.print("\n[bold]Installation details:[/bold]")
        console.print(f"  Installer: {installer_path}")
        console.print(f"  Destination: {install_dir_path}")
        if license_id:
            console.print(f"  License: {license_id}")
        if set_default:
            console.print("  Set as default: Yes")

        if dry_run:
            console.print("\n[bold cyan]Dry run mode - no changes will be made[/bold cyan]")
            console.print("\n[bold]Would perform the following actions:[/bold]")
            console.print(f"  1. Extract installer to: {install_dir_path}")
            if license_id:
                license_dir_path = get_license_dir(install_dir_path)
                console.print(f"  2. Install license to: {license_dir_path}")
            if set_default:
                config_path = get_ida_config_path()
                console.print(f"  3. Update default IDA path in: {config_path}")
            if eula:
                console.print("  4. Accept EULA")
            return

        if not auto_confirm and not Confirm.ask("\n[bold yellow]Proceed with installation?[/bold yellow]"):
            console.print("[yellow]Installation cancelled.[/yellow]")
            return

        console.print(f"[yellow]Installing {installer_path} to {install_dir_path}...[/yellow]")

        install_ida(installer_path, install_dir_path)

        if license_id:
            await get_license.callback(lid=license_id, output_dir=tmp_dir)

            # Find a file *{license_id}.hexlic in tmp_dir
            license_files = list(Path(tmp_dir).glob(f"*{license_id}.hexlic"))
            if not license_files:
                raise FileNotFoundError(f"License file matching *{license_id}.hexlic not found in {tmp_dir}")
            license_file = license_files[0].name

            license_dir_path = get_license_dir(install_dir_path)

            install_license(Path(tmp_dir) / license_file, license_dir_path)

        if set_default:
            config_path = get_ida_config_path()
            if not config_path.exists():
                console.print("[yellow]Updating configuration (default installation)...[/yellow]")
                config_path.parent.mkdir(parents=True, exist_ok=True)
                _ = config_path.write_text(json.dumps({"Paths": {"ida-install-dir": str(install_dir_path.absolute())}}))
                console.print("[grey69]Wrote default ida-config.json[/grey69]")
            else:
                # we update this without Pydantic validation to ensure we always can make the changes
                # and leave config validation to the code that requires interpretation of the file.
                doc = json.loads(config_path.read_text(encoding="utf-8"))
                if "Paths" not in doc:
                    doc["Paths"] = {}
                existing = doc["Paths"].get("ida-install-dir") or "(empty)"
                new = str(install_dir_path.absolute())
                doc["Paths"]["ida-install-dir"] = new
                _ = config_path.write_text(json.dumps(doc), encoding="utf-8")
                console.print("[grey69]Updated ida-config.json:[/grey69]")
                console.print(f"[grey69]  default install path: {existing}[/grey69]")
                console.print(f"[grey69]                     -> {new}[/grey69]")

        # this requires using ida_registry to set some keys
        # which requires idalib to be working
        # so it has to go after license and config installation
        if eula and version.product:
            if version.product in ("IDA Free", "IDA Home", "IDA Classroom"):
                # these products don't include idalib, which is used to write to the registry.
                console.print("[yellow]Skipped EULA acceptance due to product features.[/yellow]")
            else:
                # maybe its safer to have an allow-list for products with idalib
                console.print("[yellow]Accepting EULA...[/yellow]")
                try:
                    accept_eula(get_ida_path(install_dir_path))
                except RuntimeError:
                    console.print("[red]Skipped EULA acceptance due to missing idalib.[/red]")

        console.print("[green]Installation complete![/green]")

    except Exception as e:
        console.print(f"[red]Install failed: {e}[/red]")
        raise



import os
import sys
from pathlib import Path


def get_default_cache_directory() -> Path:
    if "HCLI_CACHE_DIR" in os.environ:
        return Path(os.environ["HCLI_CACHE_DIR"])

    # otherwise use XDG_CACHE_HOME
    # ref: https://github.com/mandiant/capa/issues/1212#issuecomment-1361259813
    #
    # Linux:   $XDG_CACHE_HOME/hex-rays/hcli/
    # Windows: %LOCALAPPDATA%\hex-rays\hcli\cache
    # MacOS:   ~/Library/Caches/hex-rays/hcli/

    # ref: https://stackoverflow.com/a/8220141/87207
    if sys.platform == "linux" or sys.platform == "linux2":
        directory = Path(os.environ.get("XDG_CACHE_HOME", Path.home() / ".cache" / "hex-rays" / "hcli"))
    elif sys.platform == "darwin":
        directory = Path.home() / "Library" / "Caches" / "hex-rays" / "hcli"
    elif sys.platform == "win32":
        directory = Path(os.environ["LOCALAPPDATA"]) / "hex-rays" / "hcli" / "cache"
    else:
        raise NotImplementedError(f"unsupported platform: {sys.platform}")

    directory.mkdir(parents=True, exist_ok=True)

    return directory


def validate_path_component(name: str):
    if not name or name == "." or name == "..":
        raise ValueError(f"Invalid path component: '{name}'.")

    try:
        name.encode("ascii")
    except UnicodeEncodeError:
        raise ValueError(f"Invalid path component: '{name}'. Must contain only ASCII characters")

    if "\t" in name or "\n" in name or "\r" in name:
        raise ValueError(f"Invalid path component: '{name}'. Cannot contain tabs or newlines")

    if "/" in name or "\\" in name:
        raise ValueError(f"Invalid path component: '{name}'. Cannot contain slashes")


def get_cache_directory(*key: str) -> Path:
    dir = get_default_cache_directory()
    for k in key:
        validate_path_component(k)
        dir = dir / k
    dir.mkdir(parents=True, exist_ok=True)
    return dir



"""Plugin configuration management commands."""

from __future__ import annotations

import json
import logging
import sys

import rich.table
import rich_click as click

from hcli.lib.console import console
from hcli.lib.ida import get_ida_config
from hcli.lib.ida.plugin.install import get_metadata_from_plugin_directory, get_plugin_directory
from hcli.lib.ida.plugin.settings import del_plugin_setting, get_plugin_setting, parse_setting_value, set_plugin_setting

logger = logging.getLogger(__name__)


@click.group()
@click.argument("plugin_name")
@click.pass_context
def config(ctx, plugin_name: str) -> None:
    """Manage plugin configuration settings."""
    ctx.ensure_object(dict)
    ctx.obj["config_plugin_name"] = plugin_name


@config.command()
@click.argument("key")
@click.pass_context
def get(ctx, key: str) -> None:
    """Get a plugin configuration setting."""
    plugin_name = ctx.obj["config_plugin_name"]
    try:
        value = get_plugin_setting(plugin_name, key)
        if isinstance(value, bool):
            console.print("true" if value else "false")
        else:
            console.print(value)
    except KeyError as e:
        console.print(f"[red]Error[/red]: {e}")
        raise click.Abort()
    except Exception as e:
        logger.debug("error: %s", e, exc_info=True)
        console.print(f"[red]Error[/red]: {e}")
        raise click.Abort()


@config.command()
@click.argument("key")
@click.argument("value")
@click.pass_context
def set(ctx, key: str, value: str) -> None:
    """Set a plugin configuration setting."""
    plugin_name = ctx.obj["config_plugin_name"]
    try:
        plugin_path = get_plugin_directory(plugin_name)
        metadata = get_metadata_from_plugin_directory(plugin_path)
        descr = metadata.plugin.get_setting(key)
        parsed_value = parse_setting_value(descr, value)
        set_plugin_setting(plugin_name, key, parsed_value)
        console.print(f"[green]Set[/green] {plugin_name}.{key}")
    except Exception as e:
        logger.debug("error: %s", e, exc_info=True)
        console.print(f"[red]Error[/red]: {e}")
        raise click.Abort()


@config.command(name="del")
@click.argument("key")
@click.pass_context
def del_(ctx, key: str) -> None:
    """Delete a plugin configuration setting."""
    plugin_name = ctx.obj["config_plugin_name"]
    try:
        del_plugin_setting(plugin_name, key)
        console.print(f"[green]Deleted[/green] {plugin_name}.{key}")
    except KeyError as e:
        console.print(f"[red]Error[/red]: {e}")
        raise click.Abort()
    except Exception as e:
        logger.debug("error: %s", e, exc_info=True)
        console.print(f"[red]Error[/red]: {e}")
        raise click.Abort()


@config.command()
@click.pass_context
def list(ctx) -> None:
    """List all configuration settings for a plugin."""
    plugin_name = ctx.obj["config_plugin_name"]
    try:
        plugin_path = get_plugin_directory(plugin_name)
        metadata = get_metadata_from_plugin_directory(plugin_path)

        if not metadata.plugin.settings:
            console.print(f"[grey69]No settings defined for {plugin_name}[/grey69]")
            return

        table = rich.table.Table(show_header=True, box=None)
        table.add_column("Key", style="blue")
        table.add_column("Value", style="default")
        table.add_column("Description", style="grey69")

        for setting in metadata.plugin.settings:
            try:
                value = get_plugin_setting(plugin_name, setting.key)
                value_str = "true" if value is True else ("false" if value is False else str(value))
                if setting.default is not None and value == setting.default:
                    value_str = f"{value_str} [grey69](default)[/grey69]"
            except KeyError:
                value_str = "[grey69]<not set>[/grey69]"

            description = setting.documentation or ""
            if setting.choices:
                choices_str = ", ".join(setting.choices)
                description = (
                    f"{description}\n[grey69]Choices: {choices_str}[/grey69]"
                    if description
                    else f"[grey69]Choices: {choices_str}[/grey69]"
                )

            table.add_row(setting.key, value_str, description)

        console.print(table)

    except Exception as e:
        logger.debug("error: %s", e, exc_info=True)
        console.print(f"[red]Error[/red]: {e}")
        raise click.Abort()


@config.command()
@click.pass_context
def export(ctx) -> None:
    """Export plugin configuration settings as JSON."""
    plugin_name = ctx.obj["config_plugin_name"]
    try:
        config = get_ida_config()
        if plugin_name not in config.plugins:
            console.print("{}")
            return

        plugin_config = config.plugins[plugin_name]
        console.print(json.dumps(plugin_config.settings, indent=2))

    except Exception as e:
        logger.debug("error: %s", e, exc_info=True)
        console.print(f"[red]Error[/red]: {e}")
        raise click.Abort()


@config.command(name="import")
@click.argument("json_input", required=False)
@click.pass_context
def import_(ctx, json_input: str | None) -> None:
    """Import plugin configuration settings from JSON."""
    plugin_name = ctx.obj["config_plugin_name"]
    try:
        if json_input:
            data = json.loads(json_input)
        else:
            data = json.load(sys.stdin)

        if not isinstance(data, dict):
            raise ValueError("JSON input must be an object/dict")

        for key, value in data.items():
            set_plugin_setting(plugin_name, key, value)

        console.print(f"[green]Imported {len(data)} settings for {plugin_name}[/green]")

    except json.JSONDecodeError as e:
        console.print(f"[red]Error[/red]: Invalid JSON: {e}")
        raise click.Abort()
    except Exception as e:
        logger.debug("error: %s", e, exc_info=True)
        console.print(f"[red]Error[/red]: {e}")
        raise click.Abort()



import json


class StructuredMessage:
    def __init__(self, message, /, *args, **kwargs):
        self.message = message
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        message = self.message % self.args
        return "%s <structured: %s>" % (message, json.dumps({"message": message} | self.kwargs))


m = StructuredMessage  # optional, to improve readability



from __future__ import annotations

from pathlib import Path

import rich_click as click
from rich.console import Console

from hcli.lib.config import config_store

console = Console()


@click.command()
@click.argument("name", type=str)
@click.argument("path", type=click.Path(exists=True, path_type=Path))
def add(name: str, path: Path) -> None:
    """Add a knowledge source.

    \b
    NAME: Logical name for the source
    PATH: Filesystem path to the source
    """
    path = path.expanduser().resolve()

    if not path.exists():
        console.print(f"[red]Path does not exist: {path}[/red]")
        raise click.Abort()

    sources: dict[str, str] = config_store.get_object("ke.sources", {}) or {}

    if name in sources:
        console.print(f"[yellow]Source '{name}' already exists. Use remove first to replace it.[/yellow]")
        raise click.Abort()

    # Store the absolute path as string
    sources[name] = str(path.absolute())

    # Save back to config
    config_store.set_object("ke.sources", sources)

    console.print(f"[green]Added source '{name}' pointing to '{path.absolute()}'[/green]")
